""" Class creating an AST from a formula string
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca
"""

from yappsrt import *
from formulaparser.FormulaParser import *
from myast.Number import Number
from myast.CellRef import CellRef
from myast.RangeRef import RangeRef
from myast.Function import Function

from utility.ErrorUtility import ErrorUtility
from utility.TypeCheckUtility import TypeCheckUtility
from utility.ColumnConvertUtility import ColumnConvertUtility

errHandler  = ErrorUtility.getInstance()
typeChecker = TypeCheckUtility.getInstance()
converter   = ColumnConvertUtility.getInstance()

class FormulaToAST:
  """ Class creating an AST from a formula string
  """

  def __init__(self):
    pass

  def __getCellRef(self, ref):
    """ Given a ref string (e.g. "$A1"),convert it
        to a CellRef object (e.g. CellRef(0,0,1,0))
    """
    absoluteColumn=0
    absoluteRow=0
    column=""
    row=""
    i=0

    #Checking if the column is absolute
    if ref[i]=='$':
      absoluteColumn=1
      i+=1
    #Processing the actual column, stopping when reaching the row
    if ref[i]!='#':
      while not ((ref[i]>='0' and ref[i]<='9') or ref[i]=='$' or ref[i]=='#'):
        column+=ref[i]
        i+=1
      column=converter.stringToInt(column)
    else:
      column=-1

    #Checking if the row is absolute
    if ref[i]=='$':
      absoluteRow=1
      i+=1
    #Processing the actual row
    if ref[i]!='#':
      while i<len(ref):
        row+=ref[i]
        i+=1
      row=int(row)-1
    else:
      row=-1

    return CellRef(column,
                   row,
                   absoluteColumn,
                   absoluteRow)

  def __constructAST(self, parseTree):
    """ Recursive function reducing a parse tree to an AST.
        The resulting parse tree contains no grammar artifacts,
        only the minimal amount of information.
    """
    #number, ref, rangeref are terminals: they are not
    #reduced. An appropriate ASTNode is returned.
    if parseTree.kind=="number":
      return Number(float(parseTree.strData))
    elif parseTree.kind=="ref":
      return self.__getCellRef(parseTree.strData)
    elif parseTree.kind=="rangeref":
      refs=parseTree.strData.split(":")
      return RangeRef(self.__getCellRef(refs[0]),
                      self.__getCellRef(refs[1]))
    #Next, we have the non-terminals that need to be reduced.
    elif parseTree.kind=="function":
      args=[]
      #a function node can contain 0 or 1 arg node
      if len(parseTree.children)==1:
        argNode=parseTree.children[0]
        args=[]
        #The arg node contains all the arguments
        for arg in argNode.children:
          args.append(self.__constructAST(arg))
      return Function(parseTree.strData, args)
    elif parseTree.kind=="exp":
        i=0
        #an exp can contain only one children (e.g exp->term->expterm->number)
        #thus, we shall not return a function
        if len(parseTree.children)==1:
            return self.__constructAST(parseTree.children[0])
        #otherwise, it contain list of operands and operators,
        #which need to be transformed into a function
        else:
          operands=[]
          while i<len(parseTree.children):
            #the first operand is simply reduced and added to the list
            #of arguments
            if i==0:
              operands.append(self.__constructAST(parseTree.children[i]))
              i+=2
            #if the previous argument is a +, we simply need to append the
            #reduced argument
            elif parseTree.children[i-1].kind=="add":
              operands.append(self.__constructAST(parseTree.children[i]))
              i+=2
            #otherwise, we need to append an invSum argument.
            else:
              operands.append(Function("invSum",
                                 [self.__constructAST(parseTree.children[i])]))
              i+=2
          #we return a sum function with the operands.
          return Function("sum", operands)
    #Same process as for exp node, but with mul and div.
    elif parseTree.kind=="term":
      i=0
      if len(parseTree.children)==1:
        return self.__constructAST(parseTree.children[0])
      else:
        operands=[]
        while i<len(parseTree.children):
          if i==0:
            operands.append(self.__constructAST(parseTree.children[i]))
            i+=2
          elif parseTree.children[i-1].kind=="mul":
            operands.append(self.__constructAST(parseTree.children[i]))
            i+=2
          else:
            operands.append(Function("invMul",
                                  [self.__constructAST(parseTree.children[i])]))
            i+=2
        return Function("mul", operands)
    #Similar process as before, this case is simplier: there is no inverse
    #(i.e. '-' in sum and '/' in mul)
    elif parseTree.kind=="expterm":
      i=0
      if len(parseTree.children)==1:
        return self.__constructAST(parseTree.children[0])
      else:
        operands=[]
        while i<len(parseTree.children):
          operands.append(self.__constructAST(parseTree.children[i]))
          i+=2
        return Function("exp", operands)
    #If we have a signed factor, reduce and return an unaryMinus function.
    elif parseTree.kind=="parenExp":
      exp=self.__constructAST(parseTree.children[0])
      exp.setNumParen(exp.getNumParen()+1)
      return exp
    elif parseTree.kind=="signedfactor":
      return Function("unaryMinus",
                      [self.__constructAST(parseTree.children[0])])
    else:
      raise ValueError("Wrong type for simple expression " + parseTree.kind)

  def parseToAST(self, stringToParse):
    """ Given a formula string, return its AST representation.
        Note: A tuple is returned to propagate possible errors.
        The first element of the tuple is a boolean flag set to
        true if the parse was a success. The second element is
        the actual parse tree (None on failure). Finally,
        the last element of the tuple is the error message.
        @param stringToParse: formula string
        @type stringToParse: string
        @rtype: tuple (bool, L{ast.ASTNode}, string)
    """
    P = FormulaParser(FormulaParserScanner(stringToParse))
    parseTree=wrap_error_reporter(P, 'GOAL')
    if parseTree[0]:
      return (1,self.__constructAST(parseTree[1]),parseTree[2])
    else:
      return (0,None,parseTree[2])
