""" Static Part of DSheet's graphical user interface.
    @author: Marc Provost
    @author: Hans Vangheluwe (initial prototype)
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca
"""


__author__ =  'msdl.cs.mcgill.ca'
__version__=  '0.8'

# revisions:
#
#    08/2004 Marc Provost
#      - Added Commands: Open, Save, Save as.
#      - Improved the appearance of the GUI
#
#    17/03/2004 Marc Provost
#      - Added the timer.
#
#
#    10/03/2004 Marc Provost
#     **Initial prototype heavily recoded**
#
#     added support for cell editing:
#        - text selection, cursor.
#     added support for named colums
#     Now possible to select multiple cells and apply commands on them
#     Observer pattern
#     Commands: copy/paste/undo/redo
#     Rows/Columns are variable size (not modifiable by the user for now)
#     Fonts support (not modifiable by the user for now)
#     Refactored the design:
#        - provided a clean interface to the static part of the GUI
#        - moved all behaviour code to a statechart
#
#    10/01/2004 Hans Vangheluwe
#     refactored
#
#    04/12/2002 Jean-Sebastien Bolduc
#     thoroughly cleaned up the prototype
#
#    01/03/2002 Hans Vangheluwe
#     the initial prototype

# Required modules

from   Tkinter      import *  # basic Tkinter
import tkFileDialog           # file selection (to output Postscript in)
import tkMessageBox
import tkFont
import types                  # for type checking

from utility.ErrorUtility import *
from utility.TypeCheckUtility import TypeCheckUtility
from utility.ColumnConvertUtility import ColumnConvertUtility

from subjectdata.SSheetState import SSheetState
from subjectdata.CellCoord import CellCoord
from subjectdata.CellValue import CellValue
from subjectdata.Observer import Observer

from SSGridViewStateChart import SSGridViewStateChart # the behaviour

errHandler  = ErrorUtility.getInstance()
typeChecker = TypeCheckUtility.getInstance()
converter   = ColumnConvertUtility.getInstance()

#"MS Sans Serif"
COLUMN_FONT=("helvetica", 10)
ROW_FONT=("helvetica", 10)

import time

class InvalidKeyError(Exception):
  """
  Exception indicating a non-acceptable key was pressed
  """
  pass

class SpreadsheetGridView_Static(Frame, Observer):
  """The Static part of the GUI, observing a SSheetData subject.
     @group Accessors: set*, get*
     @group Spreadsheet size Info: getWidth, getHeight
     @group Selection of cells: extendSelectedCells,  unsetSelectedCells,
                                getSelectedCells
     @group Text editing: getTextToEdit, deleteText, drawText, selectText, unselectText,
                          increaseTextSelection, getTextSelection, drawCursor,
                          moveCursor, getCursorPosition, deleteCursor, getTextSelection
     @group Commands: save, open, saveAs, copySelectedCells, cutSelectedCells,
                      pasteSelectedCells, deleteSelectedCells
     @group Undo/Redo: undo, redo, canUndo, canRedo,
                       enableUndoButton, disableUndoButton,
                       enableRedoButton, disableRedoButton
     @group Timer: setTimeWhite, setTimeBlack, increaseTimeByOne
     @group Keyboard events Utilities: is*
     @sort: set*, get*, __init__
  """

  def __init__(self,
              subject=None,
              master=None,
              title="Spreadsheet View",
              dataOffsetHor=0, dataOffsetVer=0,
              height=30, width=7,
              cellHeight=18, cellWidth=70,
              version="V"+__version__):
    """ Spreadsheet GridView's constructor

        @param subject: The subject that this object observes
        @param master: The master Tk Frame
        @param title: The title of the view (appearing in certain mgs)
        @param height: number of rows
        @param width: number of columns
        @param cellHeight: starting cell height
        @param cellWidth: starting cell width
        @type subject: L{subjectdata.Subject}
        @type master: L{Frame}
        @type title: string
        @type height: int
        @type width: int
        @type cellHeight: int
        @type cellWidth: int
    """


    #Initializing the Dynamic part of this GridView.
    #The dynamics of the SpreadsheetGridView is encoded in
    #the class SSGridViewStateChart, which was synthesized
    #by means of the SCC DCharts compiler from a DChart specification.
    #The class SpreadsheetGridView_Dynamic provides a bridge between
    #SpreadsheetGridView_Static and SSGridViewStateChart.
    self.dynamicGridView=SpreadsheetGridView_Dynamic(self)

    #In the rest of the constructor, we construct the
    #initial default GridView that will be dynamically
    #modified by the statechart.

    # Make arguments available to all methods of this class
    # Note: should really test for validity
    #       for example, height and width should be >= 1

    # the window's title
    self.__title = title

    # height in pixels of one cell
    self.__cellHeight = cellHeight

    # width in pixels of one cell
    self.__cellWidth  = cellWidth

    # number of rows
    self.__height=height

    # number of columns
    self.__width=width


    # X position of each column (each column is a canvas line)
    self.__cols=[]

    # TK tag mapped to each column (each column is a canvas line)
    self.__colsTag=[]

    # TK tag mapped to the canvas column text (i.e. A B C D ...)
    self.__colsTagIndex=[]

    # Y position of each row (each row is a canvas line)
    self.__rows=[]

    # TK tag mapped to each row (each column is a canvas line)
    self.__rowsTag=[]

    # TK tag mapped to the canvas row text (i.e. 1 2 3 4 ... )
    self.__rowsTagIndex=[]

    self.__version=version

    # Setting up the cols X position array
    deltaX=cellWidth
    for col in xrange(width+1):
      self.__cols.append(deltaX)
      deltaX+=cellWidth

    # Setting up the rows Y position array
    deltaY=cellHeight
    for row in xrange(height+1):
      self.__rows.append(deltaY)
      deltaY+=cellHeight

    # Initialize superclass
    Frame.__init__(self, master)

    # parent frame
    self.root=master
    # what to do when the window is closed
    # (event received from the Window Manager)
    self.root.protocol("WM_DELETE_WINDOW", self.wmQuit)

    # Datastructure holding the cell formulas.  In this version,
    # The observer pattern is not implemented, so we have
    # one observer per subject.
    # Here, each observer (SpreadsheetGridView) has its own subject
    self.subject=subject

    # Mapping by tuple (col,row) to each text object drawn
    # on the canvas spreadsheet (drawarea)
    self.textOnCanvas={}
    self.fontOnCanvas={}
    self.__textSelectionTag=None
    self.__currentTextSelection=None

    #Default font of the text
    self.__defaultFont=tkFont.Font(family="times", size=8)

    # Initialize packer
    self.pack(fill=BOTH, expand=1)

    # Initialize the overCell object
    # which contains the coordinates of the cell
    # the mouse is currently over.
    # This could be any cell, but (0,0) seems most reasonable
    # We make sure this is a meaningful cell (within the
    # spreadsheet) by "snapping" it into the grid boundaries
    # by means of spreadsheetBoundedCoord()
    self.__overCell = CellCoord(0,0)

    # The Tk variable overCellString
    # is used to display the coordinates of the cell the mouse is
    # currently over (modulo spreadsheet size)
    # It is updated in the callback routines for events which
    # change the overCell
    self.overCellString = StringVar()
    self.overCellString.set(str(self.__overCell))

    # The Tk variable overCellData
    # is used to display the data in the cell the mouse is currently over
    # (modulo spreadsheet size)
    # It is updated in the callback routine for events which
    # change the overCell
    self.overCellData = StringVar()
    self.overCellData.set("")

    # Tk variable mapped to the time counter label
    # Will be updated at every second by the statechart
    # describing the dynamics
    self.__timeData = StringVar()
    self.__timeData.set("00:00")

    # Initialize selectedCell to be equal to overCell
    # This could be some other cell,
    # but it seems reasonable to make this the same as overCell.
    # Note: selectedCell determines where typed in data will go.
    self.__selectedCell=CellCoord(self.__overCell.col(), self.__overCell.row())

    self.__selectedCells=None
    self.__selectedCellsTag=None

    #Tag representing the red rectangles used to indicate the selected cells
    self.__columnIndicatorTag=None
    self.__rowIndicatorTag=None

    # TK tag mapped to the overcell rectangle
    self.overCellTag=None
    # TK tag mapped to the selectedcell rectangle
    self.__selectedCellTag=None

    self.__cursorInfo=None

    self.__pivot = None

    #Used once the spreadsheet is saved as..
    self.__savedFileName=None

    # Create all widgets in the Frame
    self.__createWidgets()

  def __createWidgets(self):
    """ Function creating the GridView's widgets (at instantiation time).
    """

    self.__createPullDownMenus()

    # Frame to hold top labels and print button
    self.topFrame = Frame(master=self, background='white', border=2)
    self.topFrame.pack(side=TOP,fill=X)

    # Label displaying the content of the cell the mouse is
    # currently over.
    self.overCellContent = Label(master=self.topFrame,
                                 takefocus=0,
                                 textvariable=self.overCellData,
                                 width=20,
                                 relief="groove",
                                 background='white',
                                 foreground='blue')
    self.overCellContent.pack(side=LEFT, padx=4)

    # Label displaying the cell the mouse is currently over
    # The Tk variable self.overCellString is used for this
    self.overCellCoord = Label(master=self.topFrame,
                               takefocus=0,
                               textvariable=self.overCellString,
                               width=20,
                               relief="groove",
                               background='white',
                               foreground='blue')
    self.overCellCoord.pack(side=LEFT, padx=4)

    # Label displaying the time since the application started
    self.__timeLabel = Label(master=self.topFrame,
                           takefocus=0,
                           textvariable=self.__timeData,
                           width=20,
                           relief="groove",
                           background='white',
                           foreground='black')
    self.__timeLabel.pack(side=LEFT, padx=4)


    # Canvas (default 15cm x 8cm), stretchable
    # Scrollable region calculated from
    #   self.__height, self.__width
    #   self.__cellHeight, self.__cellWidth
    scrollHor = self.__cols[-1]
    scrollVer = self.__rows[-1]
    self.drawarea = Canvas(master=self,
                           takefocus=1,
                           width="15c", height="8c",
                           background="white",
                           scrollregion=(0, 0, scrollHor, scrollVer))

    # frame for scrollX scrollbar as well as blank corner
    self.scrollXframe = Frame(self)

    # scrollbars for drawarea
    self.scrollX = Scrollbar(self.scrollXframe, takefocus=0, orient=HORIZONTAL)
    self.scrollY = Scrollbar(self, takefocus=0, orient=VERTICAL)

    # link canvas, scrollbars, and scrolling events
    self.drawarea['xscrollcommand'] = self.scrollX.set
    self.drawarea['yscrollcommand'] = self.scrollY.set
    self.scrollX['command'] = self.drawarea.xview
    self.scrollY['command'] = self.drawarea.yview

    # a blank square for aesthetic reasons
    self.blankSquare = Canvas(self.scrollXframe, bg='lightgrey',
                       width=int(self.scrollY['width'])+2,
                       height=int(self.scrollX['width'])+2)

    # pack it all
    self.blankSquare.pack(side=RIGHT)
    self.scrollX.pack(side=LEFT, fill=X, expand=1)
    self.scrollXframe.pack(side=BOTTOM, fill=X)
    self.scrollY.pack(side=RIGHT, fill=Y)

    self.drawarea.pack(fill=BOTH, expand=1)


    # Make sure the Canvas has the focus
    # This is useful if there are other widgets around in the same
    # window such as Entry fields.
    self.drawarea.focus_set()

    self.drawarea.create_rectangle(\
      0,0,self.__cellWidth,self.__cellHeight,
      width=1, outline="grey", fill="#EBF5FF")

    self.__drawColBackground("#EBF5FF")
    self.__drawRowBackground("#EBF5FF")
    self.__drawColIndex()
    self.__drawRowIndex()
    self.__drawRows()
    self.__drawColumns()
    self.__drawOverCell()
    self.__drawSelectedCell()

    # Bind mouse events in the drawarea to callback methods that will
    # generate appropriate events in the behaviour statechart
    self.drawarea.bind("<Double-ButtonPress-1>", self.dynamicGridView.onDrawareaDoubleClick)
    self.drawarea.bind("<Motion>", self.dynamicGridView.onDrawareaMouseMove)
    self.drawarea.bind("<ButtonPress-1>", self.dynamicGridView.onDrawareaMouse1Click)
    self.drawarea.bind("<ButtonRelease-1>", self.dynamicGridView.onDrawareaMouse1Release)
    self.drawarea.bind("<Left>", self.dynamicGridView.onDrawareaLeft)
    self.drawarea.bind("<Right>", self.dynamicGridView.onDrawareaRight)
    self.drawarea.bind("<Up>", self.dynamicGridView.onDrawareaUp)
    self.drawarea.bind("<Down>", self.dynamicGridView.onDrawareaDown)
    self.drawarea.bind("<Return>", self.dynamicGridView.onDrawareaDown)
    self.drawarea.bind("<KP_Enter>", self.dynamicGridView.onDrawareaDown)
    self.drawarea.bind("<KeyPress>", self.dynamicGridView.onDrawareaKeyPressed)
    self.drawarea.bind("<Shift-KeyPress>", self.dynamicGridView.onDrawareaKeyPressed)
    self.drawarea.bind("<Control-KeyPress-c>", self.dynamicGridView.onDrawareaControlC)
    self.drawarea.bind("<Control-KeyPress-v>", self.dynamicGridView.onDrawareaControlV)
    self.drawarea.bind("<Control-KeyPress-z>", self.dynamicGridView.onDrawareaControlZ)
    self.drawarea.bind("<Control-KeyPress-y>", self.dynamicGridView.onDrawareaControlY)
    self.drawarea.bind("<Control-KeyPress-x>", self.dynamicGridView.onDrawareaControlX)
    self.drawarea.bind("<KeyPress-Shift_L>", self.dynamicGridView.onDrawareaShiftPress)
    self.drawarea.bind("<KeyRelease-Shift_L>", self.dynamicGridView.onDrawareaShiftRelease)
    self.drawarea.bind("<KeyPress-Shift_R>", self.dynamicGridView.onDrawareaShiftPress)
    self.drawarea.bind("<KeyRelease-Shift_R>", self.dynamicGridView.onDrawareaShiftRelease)


  def __createPullDownMenus(self):
    """Create the menus: File, Edit, etc.
    """

    self.menuBar = Menu(self)

    # create a File pulldown menu, and add it to the menu bar
    self.fileMenu = Menu(self.menuBar, tearoff=0)
    self.fileMenu.add_command(label="New", command=self.createNewSubject)
    self.fileMenu.add_command(label="Add gridview...", command=self.askGridViewInfo)
    self.fileMenu.add_separator()
    self.fileMenu.add_command(label="Open", command=self.open)
    self.fileMenu.add_command(label="Save",state="disabled", command=self.save)
    self.fileMenu.add_command(label="Save as...", command=self.saveAs)
    self.fileMenu.add_command(label="Print view...",
     command=self.dynamicGridView.printView)
    self.fileMenu.add_separator()
    self.fileMenu.add_command(label="Exit",command=self.wmQuit)
    self.menuBar.add_cascade(label="File", menu=self.fileMenu)

    # create an Edit pulldown menu, and add it to the menu bar
    self.editMenu = Menu(self.menuBar, tearoff=0)
    self.editMenu.add_command(label="Undo",
                              state="disabled",
                              command=self.dynamicGridView.onDrawareaControlZ,
                              accelerator="Ctrl+Z")
    self.editMenu.add_command(label="Redo",
                              state="disabled",
                              command=self.dynamicGridView.onDrawareaControlY,
                              accelerator="Ctrl+Y")
    self.editMenu.add_separator()
    self.editMenu.add_command(label="Cut",
                              state="active",
                              command=self.dynamicGridView.onDrawareaControlX,
                              accelerator="Ctrl+X")
    self.editMenu.add_command(label="Copy",
                              state="active",
                              command=self.dynamicGridView.onDrawareaControlC,
                              accelerator="Ctrl+C")
    self.editMenu.add_command(label="Paste",
                              state="active",
                              command=self.dynamicGridView.onDrawareaControlV,
                              accelerator="Ctrl+V")
    self.editMenu.add_command(label="Paste special",state="disabled")
    self.editMenu.add_separator()
    self.editMenu.add_command(label="Find...",state="disabled")
    self.editMenu.add_command(label="Replace...",state="disabled")
    self.menuBar.add_cascade(label="Edit", menu=self.editMenu)

    self.root.config(menu=self.menuBar)

  def __drawColBackground(self, color="grey"):
    """ Draws the background of the column indexes (A,B,C, ...)
        @param color: Color of the background
        @type color: string
    """
    if typeChecker.debug: typeChecker.typeCheck([color],[str])
    self.drawarea.create_rectangle(self.__cellWidth,
                                   0,
                                   self.__cols[-1],
                                   self.__cellHeight,
                                   width=0,
                                   outline="grey",
                                   fill=color)

  def __drawRowBackground(self, color="grey"):
    """ Draws the background of the row indexes (1,2,3, ...)
        @param color: Color of the background
        @type color: string
    """
    if typeChecker.debug: typeChecker.typeCheck([color],[str])
    self.drawarea.create_rectangle(0,
                                   self.__cellHeight,
                                   self.__cellWidth,
                                   self.__rows[-1],
                                   width=1,
                                   outline="grey",
                                   fill=color)

  def __indicateSelectedCell(self, color="#FFFFCC"):
    """ Indicate which cell is selected
        @param color: Color of the background
        @type color: string
    """
    if typeChecker.debug: typeChecker.typeCheck([color],[str])
    row=self.__selectedCell.row()
    col=self.__selectedCell.col()

    self.drawarea.delete(self.__columnIndicatorTag)
    self.drawarea.delete(self.__rowIndicatorTag)

    if row+1<len(self.__rows) and col+1<len(self.__cols):
      self.__columnIndicatorTag = self.drawarea.create_rectangle(
                                     self.__cols[col],
                                     0,
                                     self.__cols[col+1],
                                     self.__cellHeight,
                                     width=1,
                                     outline="grey",
                                     fill=color)
      self.drawarea.lift(self.__colsTagIndex[col])
      self.__rowIndicatorTag    = self.drawarea.create_rectangle(
                                     0,
                                     self.__rows[row],
                                     self.__cellWidth,
                                     self.__rows[row+1],
                                     width=1,
                                     outline="grey",
                                     fill=color)
      self.drawarea.lift(self.__rowsTagIndex[row])


  def __drawColIndex(self):
    """Draws the column indexes (A,B,C, ...)
    """
    self.__colsTagIndex=[]
    column=0
    for i in xrange(1,len(self.__cols)):
      t=self.drawarea.create_text(\
                        self.__cols[i-1]+(self.__cols[i]-self.__cols[i-1])*0.5,
                        self.__cellHeight*0.5,
                        font=COLUMN_FONT,
                        text=converter.intToString(column))
      self.__colsTagIndex.append(t)
      column+=1

  def __drawRowIndex(self):
    """Draws the row indexes (1,2,3, ...)
    """
    self.__rowsTagIndex=[]
    row=1
    for i in xrange(1,len(self.__rows)):
      t=self.drawarea.create_text(\
        self.__cellWidth*0.5,
        self.__rows[i-1]+(self.__rows[i]-self.__rows[i-1])*0.5,
        font=ROW_FONT,
        text=str(row))
      self.__rowsTagIndex.append(t)
      row+=1

  def __drawRows(self):
    """ Draws the horizontal lines (rows). For performance reasons,
        each cell is not explicitely represented as a rectangle;
        instead, self.__rows horizontal and self.__cols vertical lines
        are drawn
    """
    self.__rowsTag=[]
    for i in xrange(0,len(self.__rows)):
      t=self.drawarea.create_line(0,
                                  self.__rows[i],
                                  self.__cols[-1],
                                  self.__rows[i],
                                  fill="grey")
      self.__rowsTag.append(t)

  def __drawColumns(self):
    """ Draws the vertical lines (columns).  For performance reasons
        each cell is not explicitely represented as a rectangle;
        instead, self.__rows horizontal and self.__cols vertical lines
        are drawn
    """
    self.__colsTag=[]
    for i in xrange(0,len(self.__cols)):
      t=self.drawarea.create_line(self.__cols[i],
                                  0,
                                  self.__cols[i],
                                  self.__rows[-1],
                                  fill="grey")
      self.__colsTag.append(t)

  def __drawSelectedCell(self):
    """ Redraw the selected cell (tick black rectangle)
        in its new position. Called when the selected cell is
        moved using moveSelectedCell
    """
    row=self.__selectedCell.row()
    col=self.__selectedCell.col()
    self.drawarea.delete(self.__selectedCellTag)
    self.__selectedCellTag = self.drawarea.create_rectangle(\
                             self.__cols[col]+1,
                             self.__rows[row]+1,
                             self.__cols[col+1],
                             self.__rows[row+1],
                             width=2, outline="black")
    self.__indicateSelectedCell()

  def __drawSelectedCells(self, cornerA, cornerB, color="#648BD8"):
    """ Draw a rectangle surrounding a selection of cells. Order of
        the cell coordinates if not important.
        @param cornerA: First extreme of the range
        @param cornerB: Second extreme of the range
        @type cornerA: L{CellCoord}
        @type cornerB: L{CellCoord}
    """
    if typeChecker.debug: typeChecker.typeCheck([cornerA, cornerB, color],
                                                [CellCoord, CellCoord, str])
    colA=min(cornerA.col(),cornerB.col())
    rowA=min(cornerA.row(),cornerB.row())
    colB=max(cornerA.col(),cornerB.col())
    rowB=max(cornerA.row(),cornerB.row())

    X1=self.__cols[colA]
    Y1=self.__rows[rowA]
    X2=self.__cols[colB+1]
    Y2=self.__rows[rowB+1]

    self.__selectedCellsTag=self.drawarea.create_rectangle(X1+1,Y1+1,X2+1,Y2+1,width=2,outline=color,fill="#EBF5FF")
    self.drawarea.tag_lower(self.__selectedCellsTag)

  def __deleteSelectedCells(self):
    """ Deletes the current cell selection
    """
    if self.__selectedCellsTag!=None:
      self.drawarea.delete(self.__selectedCellsTag)


  def __drawOverCell(self):
    """ Redraw the selected cell (thin blue rectangle) in its new position.
        Called when the over cell is moved using moveOverCell
    """
    row=self.__overCell.row()
    col=self.__overCell.col()
    self.drawarea.delete(self.overCellTag)
    self.overCellTag = self.drawarea.create_rectangle(\
                         self.__cols[col],
                         self.__rows[row],
                         self.__cols[col+1],
                         self.__rows[row+1],
                         width=1, outline="blue")

  def __getColBoundary(self, mouseX):
    """ Get the indexes values of the two lines surrounding
        mouseX.
        @param mouseX: mouse X coordinate value
        @type mouseX: float or int
    """
    if typeChecker.debug: typeChecker.typeCheck([mouseX],[(float, int)])
    mouseX=self.drawarea.canvasx(mouseX)
    cellCol = 0
    check=0
    for i in xrange(len(self.__cols)):
      if self.__cols[i]>mouseX:
        check=1
        if i==0:
          cellCol=0
        else:
          cellCol=i-1
        break
    if check:
      return (cellCol,cellCol+1)
    else:
      return (len(self.__cols)-2,len(self.__cols)-1)

  def __getRowBoundary(self, mouseY):
    """ Get the indexes values of the two rows surrounding
        mouseY.
        @param mouseY: mouse Y coordinate value
        @type mouseY: float
    """
    if typeChecker.debug: typeChecker.typeCheck([mouseY],[(float, int)])
    mouseY=self.drawarea.canvasy(mouseY)
    cellRow = 0
    check=0
    for i in xrange(len(self.__rows)):
      if self.__rows[i]>mouseY:
        check=1
        if i==0:
          cellRow=0
        else:
          cellRow=i-1
        break
    if check:
      return (cellRow,cellRow+1)
    else:
      return (len(self.__rows)-2,len(self.__rows)-1)


  def __getCellRange(self, cornerA, cornerB):
    """ Get a list of cell coordinates contained in the specified range
        @param cornerA: First cell coordinate of the range
        @param cornerB: Second cell coordinate of the range
        @type cornerA: L{CellCoord}
        @type cornerB: L{CellCoord}
        @rtype: 2D list of L{CellCoord}
    """
    if typeChecker.debug: typeChecker.typeCheck([cornerA, cornerB],
                                                [CellCoord, CellCoord])
    cols=[]

    colA=min(cornerA.col(),cornerB.col())
    colB=max(cornerA.col(),cornerB.col())
    rowA=min(cornerA.row(),cornerB.row())
    rowB=max(cornerA.row(),cornerB.row())

    for i in xrange(rowA,rowB+1):
      rows=[]
      for j in xrange(colA, colB+1):
        rows.append(CellCoord(j,i))
      cols.append(rows)

    return cols

  def __getCharPos(self, cellCoord, event):
    """ Get the char closest to event in cellCoost
        @param cellCoord: cell coordinate where the text is
        @param event: user interaction event
        @type cellCoord: L{CellCoord}
        @type event: Event
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord, event],
                                                [CellCoord, (type(None), Event)])
    event.x=self.drawarea.canvasx(event.x)
    event.y=self.drawarea.canvasy(event.y)
    if not self.textOnCanvas.has_key(cellCoord):
      return 0

    position=0
    textTag=self.textOnCanvas[cellCoord]
    textFont=self.fontOnCanvas[cellCoord]
    text=self.drawarea.itemcget(textTag, "text")

    if len(text)==0:
      return 0

    col0=self.__cols[cellCoord.col()]
    col1=self.__cols[cellCoord.col()+1]
    row0=self.__rows[cellCoord.row()]
    row1=self.__rows[cellCoord.row()+1]

    if event!=None:
      x=event.x-col0
      y=event.y-row0
      if x>=textFont.measure(text):
        position=len(text)
      elif x<textFont.measure(text[0]):
        position=0
      else:
        for i in xrange(len(text)):
          if textFont.measure(text[:i])<=x and textFont.measure(text[:i+1])>=x:
            position=i
            break

    return position

  def __setSelectedCells(self, cornerA, cornerB):
    """ Delete the current cell selection, draw a new cell
        selection
        @param cornerA: first cell coordinate of the range
        @param cornerB: second cell coordinate of the range
        @type cornerA: L{CellCoord}
        @type cornerB: L{CellCoord}
    """
    if typeChecker.debug: typeChecker.typeCheck([cornerA, cornerB],
                                                [CellCoord, CellCoord])
    self.__deleteSelectedCells()
    self.__drawSelectedCells(cornerA, cornerB)
    self.__selectedCells=self.__getCellRange(cornerA,cornerB)
    if self.__pivot == None:
      self.__pivot = self.getSelectedCell()


  #**************************************************************************
  #
  # Next follows the interface to the gridview.  The following methods are
  # called from the statechart to express the dynamic behavior of the
  # spreadsheet.  For instance, one can move the selected cell (black tick
  # rectange) when the mouse is clicked using the moveSelectedCellTo method
  #
  # *************************************************************************

  def saveAs(self):
    """ Ask for a filename and save the spreadsheet.
        Currently not saving any graphical information.
        Using built-in subject export()
    """
    f=tkFileDialog.asksaveasfilename(defaultextension=".ds",
                                     filetypes=[("DSheet", ".ds")])
    if not self.subject.exportToFile(f):
        tkMessageBox.showwarning(
            "Save as...",
            "Unable to create the file"
        )
    else:
      self.__savedFileName=f
      self.fileMenu.entryconfig(index="Save", state="active")

  def save(self):
    """ Save the file using the previously specified filename.
        Disabled if no filename were specified yet.
    """
    if not self.subject.exportToFile(self.__savedFileName):
        tkMessageBox.showwarning(
            "Save",
            "Unable to create the file"
        )

  def open(self):
    """ Load spreadsheet's data from a .ds file.
    """
    f=tkFileDialog.askopenfilename(defaultextension=".ds",
                                   filetypes=[("DSheet", ".ds")])
    if not self.subject.importFromFile(f):
        tkMessageBox.showwarning(
            "Open",
            "Unable to load the file"
        )
    else:
      self.__savedFileName=f
      self.fileMenu.entryconfig(index="Save", state="active")


  def getSelectedCell(self):
    """ Return the currently selected cell.
        @rtype: L{CellCoord}
    """
    return self.__selectedCell

  def getOverCell(self):
    """ Return the cell the mouse is currently over.
        @rtype: L{CellCoord}
    """
    return self.__overCell

  def getTextToEdit(self, cellCoord):
    """ Querying the subject for the text to edit in the specified cell.
        i.e. The text to edit is usually a formula, if there is no formula,
        it is a string value.
        @param cellCoord: Specified cell
        @type cellCoord: L{CellCoord}
        @rtype: string
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord],[CellCoord])
    state=self.subject.getState(cellCoord,cellCoord)
    if state.getCellFormula(cellCoord)!="":
      return state.getCellFormula(cellCoord)
    else:
      cellValue=state.getCellValue(cellCoord)
      if cellValue.isString():
        return cellValue.getString()
      else:
        return ""



  def setOverCellDisplay(self, string):
    """ Update the display of the overcell (where the
        current viewed formula is displayed).
        Used for instance when the mouse is moved to
        display the formula expression of the cell under the mouse.
    """
    if typeChecker.debug: typeChecker.typeCheck([string],[str])
    self.overCellData.set(string)

  def setColRowDisplay(self, string):
    """ Update the current row/column display.
        Used for instance when the mouse is moved to
        update the current row/column display of the cell
        under the mouse.
    """
    if typeChecker.debug: typeChecker.typeCheck([string],[str])
    self.overCellString.set(string)


  #Other accessors methods to the class' private fields:

  def getHeight(self):
    """ Returns the height of the canvas (number of rows)
    """
    return self.__height

  def getWidth(self):
    """ Returns the width of the canvas (number of columns)
    """
    return self.__width

  def increaseTimeByOne(self):
    """ Increases the time counter string by 1 second.
        Note that the time counter does not include hours
        (only "MM:SS"). The counter is reset to "00:00" after "59:59".
        Note: A more efficient implementation would keep tsec and tmin
        as attributes rather than converting from the __timeData string
        every second.
    """
    t=self.__timeData.get()

    tsec=int(t[3:])

    if tsec<59:
      tsec+=1
      if tsec<10:
        self.__timeData.set(t[0:3]+"0"+str(tsec))
      else:
        self.__timeData.set(t[0:3]+str(tsec))
    else:
      tmin=int(t[0:2])
      if tmin<59:
        tmin+=1
        tsec=0
        if tmin<10:
          self.__timeData.set("0"+str(tmin)+":00")
        else:
          self.__timeData.set(str(tmin)+":00")
      else:
        self.__timeData.set("00:00")

  def setTimeBlack(self):
    """ Change the foreground/background of the time counter to white/black
        (reverse video)
    """
    self.__timeLabel.config(fg="white")
    self.__timeLabel.config(bg="black")

  def setTimeWhite(self):
    """ Change the foreground/background of the time counter to black/white
    """
    self.__timeLabel.config(fg="black")
    self.__timeLabel.config(bg="white")

  def extendSelectedCells(self, cellCoord):
    """ Extend the current cell selection to encapsulate
        cellCoord.
        @param cellCoord: Cell coordinate to include in the selection
        @type cellCoord: CellCoord
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord],[CellCoord])

    selectedCell=self.getSelectedCell()

    if self.__selectedCells == None:
      self.__setSelectedCells(selectedCell, selectedCell)

    cornerACol = min(self.__pivot.col(),
                     cellCoord.col())
    cornerARow = min(self.__pivot.row(),
                     cellCoord.row())
    cornerBCol = max(self.__pivot.col(),
                     cellCoord.col())
    cornerBRow = max(self.__pivot.row(),
                     cellCoord.row())

    self.__setSelectedCells(CellCoord(cornerACol, cornerARow),
                          CellCoord(cornerBCol, cornerBRow))

  def unsetSelectedCells(self):
    """ Delete the current cell selection
    """
    self.__deleteSelectedCells()
    self.__selectedCells=None
    self.__pivot = None

  def deleteSelectedCells(self):
    """ Delete the contents of the selected cells
    """
    cells=None
    cells=self.getSelectedCells()

    #creating a newState consisting of empty cells
    newState=SSheetState()
    for row in cells:
      for cell in row:
        newState.setCellFormula(cell,"")

    self.subject.setState(newState)

  def copySelectedCells(self):
    """ Copy the selected cells
    """
    if self.__selectedCells!=None:
      self.subject.copy(self.__selectedCells[0][0],
                        self.__selectedCells[-1][-1])
    else:
      self.subject.copy(self.__selectedCell,
                        self.__selectedCell)

  def cutSelectedCells(self):
    """ Cut the selected cells
    """
    self.copySelectedCells()
    self.deleteSelectedCells()

  def undo(self):
    """ Undo the previous command
    """
    self.subject.undo()

  def canUndo(self):
    """ Returns true if the subject can undo
        @rtype: bool
    """
    return self.subject.canUndo()

  def redo(self):
    """ Redo the latest undone command
    """
    self.subject.redo()

  def canRedo(self):
    """ Returns true if the subject can redo
        @rtype: bool
    """
    return self.subject.canRedo()

  def enableUndoButton(self):
    """ Enables the undo button
    """
    self.editMenu.entryconfig(index="Undo", state="active")

  def disableUndoButton(self):
    """ Disables the undo button
    """
    self.editMenu.entryconfig(index="Undo", state="disabled")

  def enableRedoButton(self):
    """ Enables the undo button
    """
    self.editMenu.entryconfig(index="Redo", state="active")

  def disableRedoButton(self):
    """ Disables the redo button
    """
    self.editMenu.entryconfig(index="Redo", state="disabled")

  def pasteSelectedCells(self):
    """ Paste the cells into the current selection
    """
    cellCoordA = self.getSelectedCell()
    cellCoordB = self.getSelectedCell()
    self.subject.paste(cellCoordA, cellCoordB)

  def getSelectedCells(self):
    """ Returns a two dimensional list containing the selected cells
        @rtype: 2D list of L{CellCoord}
    """
    import copy
    if self.__selectedCells!=None:
      return copy.copy(self.__selectedCells)
    else:
      return [[self.__selectedCell]]


  def moveSelectedCellTo(self, cellCoord):
    """ Move and redraw the selected cell
        to the specified coordinate.
        @param cellCoord: the coordinate
        @type cellCoord: L{CellCoord}
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord], [CellCoord])
    col = cellCoord.col()
    row = cellCoord.row()
    if col<0:
      col=0
    elif col>=self.__width:
      col=self.__width-1
    if row<0:
      row=0
    elif row>=self.__height:
      row=self.__height-1

    self.__selectedCell=CellCoord(col,row)
    self.__drawSelectedCell()

  def moveOverCellTo(self, cellCoord):
    """ Move and redraw the over cell
        to the specified coordinate.
        @param cellCoord: the cell coordinate
        @type cellCoord: L{CellCoord}
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord], [CellCoord])
    col = cellCoord.col()
    row = cellCoord.row()
    if col<0:
      col=0
    elif col>=self.__width:
      col=self.__width-1
    if row<0:
      row=0
    elif row>=self.__height:
      row=self.__height-1

    self.__overCell=CellCoord(col,row)
    self.__drawOverCell()

  def deleteText(self, cellCoord):
    """ Delete the text drawn at cellCoord
        @param cellCoord: Cell coordinate of the deleted text
        @type cellCoord: L{CellCoord}
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord],[CellCoord])
    if self.textOnCanvas.has_key(cellCoord):
      self.drawarea.delete(self.textOnCanvas[cellCoord])
      del self.textOnCanvas[cellCoord]
      del self.fontOnCanvas[cellCoord]

  def drawText(self,cellCoord,text, centered=0, font=None):
    """ Draws text at a given cell in the spreadsheet.

        NOTE: The canvas text objects are not the data holder ! They must be
        kept IN SYNCH with the subject, which is the datastructure maintaining
        the cell formulas. The text objects can also contain data that is
        currently being edited and not yet submitted to the subject
        (see acceptData).
        @param cellCoord: cell coordinate where to draw the text
        @param text: text to draw
        @param centered: flag specifying whether the text is centered
        @param font: font of the text
        @type cellCoord: L{CellCoord}
        @type text: string
        @type centered: bool
        @type font: Tuple (see Tkinter doc)
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord, text, centered, font],
                                                [CellCoord, str, int, (type(None), tkFont.Font)])
    col=cellCoord.col()
    row=cellCoord.row()
    if self.textOnCanvas.has_key(cellCoord):
      tag=self.textOnCanvas[cellCoord]
      self.drawarea.delete(tag)
    if not centered:
      deltaX=self.__cols[col]+1
    else:
      deltaX=self.__cols[col]+(self.__cols[col+1]-self.__cols[col])*0.5
    deltaY=self.__rows[row]+(self.__rows[row+1]-self.__rows[row])*0.5
    if font==None:
      if self.fontOnCanvas.has_key(cellCoord):
        font=self.fontOnCanvas[cellCoord]
      else:
        font=self.__defaultFont
    self.fontOnCanvas[cellCoord]=font

    tag=None
    if not centered:
      tag=self.drawarea.create_text(deltaX,
                                    deltaY,
                                    text=text,
                                    font=font,
                                    anchor="w",
                                    justify=CENTER)
    else:
      tag=self.drawarea.create_text(deltaX,
                                    deltaY,
                                    text=text,
                                    font=font,
                                    justify=CENTER)
    self.textOnCanvas[cellCoord]=tag


  def selectText(self, cellCoord, charRange):
    """ Select text in a specified range.
        @param cellCoord: cell coordinate where to select text
        @param charRange: range of chars to select
        @type cellCoord: L{CellCoord}
        @type charRange: tuple (int, int)
    """
    if typeChecker.debug:
      typeChecker.typeCheck([cellCoord, charRange],
                            [CellCoord, tuple])
      for elem in charRange:
        typeChecker.typeCheck([elem], [int])

    if self.__textSelectionTag !=None:
      self.drawarea.delete(self.__textSelectionTag)
    col0=self.__cols[cellCoord.col()]
    col1=self.__cols[cellCoord.col()+1]
    row0=self.__rows[cellCoord.row()]
    row1=self.__rows[cellCoord.row()+1]

    Y0=row0
    Y1=row1
    if not self.textOnCanvas.has_key(cellCoord):
      X0=col0
      X1=col1
    else:
      textTag=self.textOnCanvas[cellCoord]
      textFont=self.fontOnCanvas[cellCoord]
      text=self.drawarea.itemcget(textTag, "text")

      if charRange[0] < 0 or charRange[1] > len(text):
        X0=col0
        X1=col1
        charRange=(0,len(text))
      else:
        X0=col0+textFont.measure(text[:charRange[0]])
        X1=col0+textFont.measure(text[:charRange[1]])

    self.__textSelectionTag=self.drawarea.create_rectangle(X0,Y0,X1,Y1,fill="#FFFF99",width=0)

    if self.textOnCanvas.has_key(cellCoord):
      textTag=self.textOnCanvas[cellCoord]
      self.drawarea.tag_lower(self.__textSelectionTag,textTag)

    self.__currentTextSelection=(cellCoord,charRange)

  def unselectText(self):
    """ Unselect current selected text
    """
    if self.__textSelectionTag !=None:
      self.drawarea.delete(self.__textSelectionTag)
      self.__currentTextSelection=None
      self.__textSelectionTag=None

  def increaseTextSelection(self, cellCoord, deltaX=0, event=None):
    """ Increase the text selection by deltaX or up to an event position
        @param cellCoord: CellCoordinate where to select the text
        @param deltaX: increase value (can be negative to reduce selection)
        @param event: modify the range up to the event location
        @type cellCoord: CellCoord
        @type deltaX: int or float
        @type event: L{Event}
        Note: if event is specified, it overrides deltaX
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord, deltaX, event],
                                                [CellCoord, (int,float), (type(None), Event)])
    if self.__cursorInfo==None:
      return

    cursorInfo=self.__cursorInfo

    if event!=None:
      position=self.__getCharPos(cellCoord, event)
      deltaX=position-cursorInfo[4]

    if deltaX<0:
      if self.__currentTextSelection == None:
        self.selectText(cellCoord,
                           (cursorInfo[4]+deltaX,cursorInfo[4]))
      else:
        if cursorInfo[4]==self.__currentTextSelection[1][0]:
          self.selectText(cellCoord,
                             (self.__currentTextSelection[1][0]+deltaX,
                              self.__currentTextSelection[1][1]))
        else:
          self.selectText(cellCoord,
                             (self.__currentTextSelection[1][0],
                              self.__currentTextSelection[1][1]+deltaX))

    else:
      if self.__currentTextSelection == None:
        self.selectText(cellCoord,
                           (cursorInfo[4],cursorInfo[4]+deltaX))
      else:
        if cursorInfo[4]==self.__currentTextSelection[1][1]:
          self.selectText(cellCoord,
                             (self.__currentTextSelection[1][0],
                              self.__currentTextSelection[1][1]+deltaX))
        else:
          self.selectText(cellCoord,
                             (self.__currentTextSelection[1][0]+deltaX,
                              self.__currentTextSelection[1][1]))

    self.moveCursor(deltaX)


  def getTextSelection(self):
    """ Return a tuple (cell coord, char range) describing the current
        text selection.
        @rtype: tuple (L{CellCoord}, tuple (int, int))
    """
    return self.__currentTextSelection


  def drawCursor(self, cellCoord, position=0, event=None):
    """ Animate a cursor with respect to the text that is in cellCoord.
        Only one cursor is allowed at a given time in the spreadsheet.
        If one wants to create a cursor in another cell, ``deleteCursor''
        must be called first.  If drawCursor is called two times in a row
        it will become impossible to delete one of the cursors.
        The pattern ``deleteCursor'' & ``drawCursor'' needs to be used
        each time the current edited text is modified with drawText to move
        the cursor accordingly.(see drawText). Note: default position value is 0.
        If an event is specified, position is ignored.
        @param cellCoord: The cell coordinate where to draw the cursor
        @param position: The character index where to draw the cursor (optional)
        @param event: the event where to draw the cursor
        @type cellCoord: L{CellCoord}
        @type position: int
        @type event: event
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord, position, event],
                                                [CellCoord, int, (type(None), Event)])
    textTag=self.textOnCanvas[cellCoord]
    textFont=self.fontOnCanvas[cellCoord]
    text=self.drawarea.itemcget(textTag, "text")

    if position<0:
      position=0
    if position>len(text):
      position=len(text)


    col0=self.__cols[cellCoord.col()]
    col1=self.__cols[cellCoord.col()+1]
    row0=self.__rows[cellCoord.row()]
    row1=self.__rows[cellCoord.row()+1]

    if event!=None:
      position=self.__getCharPos(cellCoord,event)

    deltaX=textFont.measure(text[0:position])
    if deltaX==0: deltaX=1

    X1=col0+deltaX+1
    Y1=row0+1
    X2=col0+deltaX+1
    Y2=row1-1

    tag=self.drawarea.create_line(X1,Y1,X2,Y2)
    deleteEvent=self.root.after(500,self.drawarea.delete,tag)
    creationEvent=self.root.after(1000,self.drawCursor,cellCoord,position)
    self.__cursorInfo=(tag,deleteEvent,creationEvent,cellCoord,position)

  def deleteCursor(self):
    """ Delete the current animated cursor.
        Is used in conjunction with ``drawCursor'' and
        drawText to move the cursor within the same edited cell.
    """
    if self.__cursorInfo!=None:
      self.drawarea.delete(self.__cursorInfo[0])
      self.root.after_cancel(self.__cursorInfo[1])
      self.root.after_cancel(self.__cursorInfo[2])
      self.__cursorInfo=None

  def moveCursor(self, deltaX):
    """ Move the cursor by a specified number of characters
        @param deltaX: number of characters (can be negative to move left)
        @type deltaX: float or int
    """
    if typeChecker.debug: typeChecker.typeCheck([deltaX],[(int,float)])
    if self.__cursorInfo==None:
      return
    position=self.__cursorInfo[4]
    position+=deltaX
    cellCoord=self.__cursorInfo[3]

    self.deleteCursor()
    self.drawCursor(cellCoord,position)

  def getCursorPosition(self):
    """ Returns the cursor position
        @rtype: int
    """
    if self.__cursorInfo==None:
      return 0
    return self.__cursorInfo[4]

  def cellCoords(self, mouseX, mouseY):
    """ Utility method to convert mouse position (event.x, event.y)
        into cell coordinates (row, col)
        @param mouseX: mouse X coordinate
        @param mouseY: mouse Y coordinate
        @type mouseX: float
        @type mouseY: float
        @rtype: L{CellCoord}
    """
    if typeChecker.debug: typeChecker.typeCheck([mouseX, mouseY], [int, int])
    cellRow=self.__getRowBoundary(mouseY)[0]
    cellCol=self.__getColBoundary(mouseX)[0]
    return CellCoord(cellCol, cellRow)

  def acceptData(self,cell,data=""):
      """ Commit data with respect to a particular cell in the subject.
          This method either succeeds or fails depending on whether there
          is a parse error when processing the formula.
          The data is NOT commited if the method  returns false, it is
          commited otherwise.  A message is printed to stout if there is
          a parse error.
          NOTE: Before, a tkMessageBox was used to display the error,
          but creating such an object from SVM freezes the system
          (GOD knows why: SVM is using threads and Tkinter does strange
          things with threads).
          @param cell: the cell coord where data is commited
          @param data: the data which is commited
          @type cell: L{CellCoord}
          @type data: string
          @rtype: bool
      """

      if len(data)>0:
        if data[0]!='=':
          try:
            data=float(data)
          except Exception,e:
            pass
          if type(data)==types.FloatType:
            data="="+str(data)

      state=SSheetState()
      if len(data)>0 and data[0]=='=':
        state.setCellFormula(cell,data)
      else:
        value=CellValue(valueString=data)
        state.setCellValue(cell,value)

      errTuple = self.subject.setState(state)
      if not errTuple[0]:
        #import tkMessageBox
        #tkMessageBox.showwarning(
        #    "Parse Error",
        #    "There is an error in the formula:\n\n"+str(errTuple[1])
        #)
        print "A parse error was found in the formula"
        print errTuple[1]
        return False
      else:
        return True

  def notify(self):
    """ Method called by the subject to notify observers that
        they need to update themselves.
    """
    before = time.clock()
    minCell=CellCoord(0,0)
    maxCell=CellCoord(self.__width-1,self.__height-1)
    state=self.subject.getState(minCell,maxCell)
    for cellCoord in state.getCoords():
      value=state.getCellValue(cellCoord)
      if value.isFloat():
        self.drawText(cellCoord,str(value.getFloat()),centered=1)
      elif value.isString():
        self.drawText(cellCoord,value.getString(), centered=1)
      else:
        self.drawText(cellCoord,"",centered=1)
    for cellCoord in self.textOnCanvas.keys():
      if not state.hasCell(cellCoord):
        self.deleteText(cellCoord)
    after = time.clock()
    #print "update took: ", str(after-before)

  def addGridView(self):
    """ Method called to create a new gridview (observing the same
        subject as this object)
    """
    # TODO: check that Strictly Positive Integers were entered
    rows     = int(self.rowsEntry.get())
    columns  = int(self.colsEntry.get())

    # get cellHeight and cellWidth from Entry field
    # TODO: check that Strictly Positive Integers were entered
    cellHeight = int(self.cellHeightEntry.get())
    cellWidth  = int(self.cellWidthEntry.get())
    self.gridView.destroy()

    self.subject.getSubjectManager().addGridViewObserver(self.subject, rows=rows,
                                                         columns=columns,
                                                         cellHeight=cellHeight,
                                                         cellWidth=cellWidth)

  def createNewSubject(self):
    """ Method called to create a new subject. An observer
        is automatically attached to the newly created subject
    """
    manager=self.subject.getSubjectManager()
    manager.addGridViewObserver(manager.createSubject())

  def askGridViewInfo(self):
    """ Method used to query the user about the size of the new gridview.
        (information to be passed to addGridView)
    """
    self.gridView=Toplevel(self.root)

    # Rows Label
    Label(master=self.gridView, text="rows:").pack(side=LEFT)

    # Entry field to enter the number of rows
    self.rowsEntry = Entry(master=self.gridView, width=4)
    self.rowsEntry.insert(0, "100")
    self.rowsEntry.pack(side=LEFT)

    # Columns Label
    Label(master=self.gridView, text="columns:").pack(side=LEFT)

    # Entry field to enter the number of columns
    self.colsEntry = Entry(master=self.gridView, width=4)
    self.colsEntry.insert(0, "50")
    self.colsEntry.pack(side=LEFT)

    # cellHeight Label
    Label(master=self.gridView, text="cellHeight:").pack(side=LEFT)

    # Entry field to enter the cellHeight
    self.cellHeightEntry = Entry(master=self.gridView, width=4)
    self.cellHeightEntry.insert(0, "18")
    self.cellHeightEntry.pack(side=LEFT)

    # cellWidth Label
    Label(master=self.gridView, text="cellWidth:").pack(side=LEFT)

    # Entry field to enter the cellWidth
    self.cellWidthEntry = Entry(master=self.gridView, width=4)
    self.cellWidthEntry.insert(0, "50")
    self.cellWidthEntry.pack(side=LEFT)

    # Button to add a grid spreadsheet view
    self.gridViewButton = Button(master=self.gridView,
                            text="Add GridView",
                            command=self.addGridView)
    self.gridViewButton.pack(side=RIGHT)


  # The two following functions were supposed to be called from the statechart,
  # but due to the fact that SVM uses threads, creating dialogs such as
  # asksaveasfilename hangs the interpreter.  Thus, this behavior was
  # left in this class.  (Note that button press events are bound to
  #  these 2 functions).

  def printView(self):
    """ Method used to print the spreadsheet to .ps files
    """
    printFilename = \
     tkFileDialog.asksaveasfilename(filetypes=[("Postscript files", "*.ps")])
    bbox=self.drawarea.bbox(ALL)
    self.drawarea.postscript(file=printFilename,
                             width=bbox[2]+10,
                 height=bbox[3]+10)

  def wmQuit(self):
    print self.__title + " was closed (by Window Manager)"
    self.root.destroy()
    self.subject.detach(self)

# The SpreadsheetGridView_Dynamic class ``bridges'' the
# + non-behaviour information encoded in the class SSGridview and
# + the behaviour information modelled in the statechart
#   (encoded in the class SSGridViewStateChart generated from the
#    statechart).
# GUI events are bound to the methods on...() of the
# SpreadsheetGridView_Dynamic class.
# These methods in turns trigger an event in
# the statechart.  Then, the statechart reacts to the event
# by calling (back) appropriate functions of the SSGridView class
# to execute the correct behavior.

class SpreadsheetGridView_Dynamic:
  """ The layer between the static part of the GUI and the
      statechart executing its behavior.
  """

  def __init__(self,staticGV):
    self.statechart=SSGridViewStateChart()
    self.statechart.initModel()
    self.statechart.event("start",staticGV)

  # All is... methods return Boolean True (1) or False (0)

  def isAlphaChar(self, keysym):
   """Helper method to check if a keypress is alphachar.
      @param keysym: tkinter key identifier
      @rtype: bool
   """
   if typeChecker.debug: typeChecker.typeCheck([keysym], [str])
   return ( (len(keysym) == 1) and
            (('a' <= keysym <= 'z') or
             ('A' <= keysym <= 'Z')) )

  def isNonAlphaChar(self, keysym):
   """ Helper method to check if a keypress is non alpha char.
       Note: <NonAlphaChar> in { /~!@#$%^&*()-_+;:'",<>?[]}
       @param keysym: tkinter key identifier
       @rtype: bool
   """
   if typeChecker.debug: typeChecker.typeCheck([keysym], [str])
   nonAlphaChars = ['slash',         # /
                    'asciitilde',    # ~
                    'exclam',        # !
                    'at',            # @
                    'numbersign',    # #
                    'dollar',        # $
                    'percent',       # %
                    'asciicircum',   # ^
                    'ampersand',     # &
                    'asterisk',      # *
                    'parenleft',     # (
                    'parenright',    # )
                    'minus',         # -
                    'underscore',    # _
                    'plus',          # +
                    'semicolon',     # ;
                    'colon',         # :
                    'apostrophe',    # '
                    'quotedbl',      # "
                    'comma',         # ,
                    'less',          # <
                    'greater',       # >
                    'question',      # ?
                    'bracketleft',   # [
                    'bracketright']  # ]
   return keysym in nonAlphaChars

  def isDigitChar(self, keysym):
   """ Helper method to check if a keypress is a digit.
       <DigitChar> in the range 0 - 9
       @param keysym: tkinter key identifier
       @rtype: bool
   """
   if typeChecker.debug: typeChecker.typeCheck([keysym], [str])
   return ('0' <= keysym <=  '9')

  def isPeriod(self, keysym):
   """ Helper method to check if a keypress is a period.
       @param keysym: tkinter key identifier
       @rtype: bool
   """
   if typeChecker.debug: typeChecker.typeCheck([keysym], [str])
   return (keysym == 'period')

  def isEquals(self, keysym):
   """ Helper method to check if a keypress is an equal.
       @param keysym: tkinter key identifier
       @rtype: bool
   """
   return (keysym == 'equal')

  def isBackSpace(self, keysym):
    return keysym=='BackSpace'

  def isDelete(self, keysym):
    return keysym=='Delete'

  def isEditKey(self, keysym):
    return self.isPeriod(keysym) or\
           self.isEquals(keysym) or\
           self.isDigitChar(keysym) or\
           self.isNonAlphaChar(keysym) or\
           self.isAlphaChar(keysym)



  #binding with the statechart:

  def onDrawareaMouseMove(self, event):
    self.statechart.event("mouseMove",event)

  def onDrawareaMouse1Click(self, event):
    self.statechart.event("mouseClick",event)

  def onDrawareaMouse1Release(self, event):
    self.statechart.event("mouseRelease",event)

  def onDrawareaLeft(self, event):
    self.statechart.event("leftKey",event)

  def onDrawareaRight(self, event):
    self.statechart.event("rightKey",event)

  def onDrawareaUp(self, event):
    self.statechart.event("upKey",event)

  def onDrawareaDown(self, event):
    self.statechart.event("downKey",event)

  def onDrawareaKeyPressed(self, event):
    if self.isEditKey(event.keysym):
      self.statechart.event("editKey",event)
    elif self.isDelete(event.keysym):
      self.statechart.event("deleteKey", event)
    elif self.isBackSpace(event.keysym):
      self.statechart.event("backSpaceKey", event)

  def onDrawareaControlC(self, event=None):
    self.statechart.event("controlC",event)

  def onDrawareaControlV(self, event=None):
    self.statechart.event("controlV",event)

  def onDrawareaControlZ(self, event=None):
    self.statechart.event("controlZ", event)

  def onDrawareaControlY(self, event=None):
    self.statechart.event("controlY", event)

  def onDrawareaControlX(self, event=None):
    self.statechart.event("controlX", event)

  def onDrawareaDoubleClick(self, event):
    self.statechart.event("doubleClick",event)

  def onDrawareaShiftPress(self, event=None):
    self.statechart.event("shiftPress")

  def onDrawareaShiftRelease(self, event=None):
    self.statechart.event("shiftRelease")

  def printView(self):
    self.statechart.event("print",event)

  def stop(self):
    self.statechart.event("quit",event)
