""" Cell Reference node of the AST.
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca    
"""

from ASTVisitor import ASTVisitor
from ASTNode import ASTNode
from utility.TypeCheckUtility import TypeCheckUtility
typeChecker=TypeCheckUtility.getInstance()

class CellRef(ASTNode):
  """ Encapsulates an (absolute or relative) reference 
      @group Accessors: set*, get*, isAbs*
      @sort: __init__, set*, get*, isAbs* 
  """

  def __init__(self, column, row, isAbsCol, isAbsRow, numParen=0):
    """ CellRef's constructor
        @param column: The column number of the reference
        @param row: The row number of the reference
        @param isAbsCol: If true, the column is absolute. Otherwise: relative.
        @param isAbsRow: If true, the row is absolute. Otherwise: relative.
        @type column: int
        @type row: int
        @type isAbsCol: bool
        @type isAbsRow: bool      
    """  
    if typeChecker.debug: typeChecker.typeCheck([column,row,isAbsCol,isAbsRow],
                                                [int,int,int,int])   
    self.__column=column
    self.__row=row
    self.__isAbsCol=isAbsCol
    self.__isAbsRow=isAbsRow
    #Number of parenthesis around
    self.__numParen=numParen    

  #Next follows accessors methods to return private fields
    
  def getRow(self):
    """ Return the row value
        @rtype: int
    """
    return self.__row
    
  def getColumn(self):
    """ Return the column value
        @rtype: int
    """
    return self.__column

  def isAbsRow(self):
    """ Return true if the row is absolute
        @rtype: bool
    """
    return self.__isAbsRow

  def isAbsColumn(self):
    """ Return true if the column is absolute
        @rtype: bool
    """
    return self.__isAbsCol

  def getNumParen(self):
    """ Returns the number of parenthesis around this ASTNode
        @rtype: int
    """
    return self.__numParen        

  def setRow(self,row):
    """ Set the row value of this CellRef
        @param row: row value
        @type row: int      
    """  
    if typeChecker.debug: typeChecker.typeCheck([row], [int])
    self.__row=row

  def setColumn(self,column):
    """ Set the column value of this CellRef
        @param column: column value
        @type column: int      
    """   
    if typeChecker.debug: typeChecker.typeCheck([column], [int])
    self.__column=column
  
  def setIsAbsRow(self,absRow):
    """ Set the row to be absolute or not
        @param absRow: flag indicating absoluteness
        @type absRow: bool      
    """    
    if typeChecker.debug: typeChecker.typeCheck([absRow], [int]) 
    self.__isAbsRow=absRow
  
  def setIsAbsColumn(self,absCol):
    """ Set the column to be absolute or not
        @param absCol: flag indicating absoluteness
        @type absCol: bool      
    """ 
    if typeChecker.debug: typeChecker.typeCheck([absCol], [int])    
    self.__isAbsCol=absCol
    
  def setNumParen(self, numParen):
    """ Set the number of parenthesis around this number
        @param numParen: number of parenthesis
        @type numParen: int
    """
    self.__numParen=numParen    

  def accept(self,visitor):
    if typeChecker.debug: typeChecker.typeCheck([visitor], [ASTVisitor])
    visitor.visitCellRef(self)

  def clone(self):
    """ Returns a copy of this object
        @rtype: L{CellRef}
    """   
    return CellRef(self.__column,
                   self.__row,
                   self.__isAbsCol,
                   self.__isAbsRow,
                   self.__numParen)

  def equals(self, anObject):
    return isinstance(anObject,CellRef) and\
           anObject.getColumn()==self.__column and\
           anObject.getRow()==self.__row and\
           anObject.isAbsColumn()==self.__isAbsCol and\
           anObject.isAbsRow()==self.__isAbsRow and\
           anObject.getNumParen()==self.__numParen 