""" Function node of the AST.
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca    
"""

from ASTVisitor import ASTVisitor
from ASTNode import ASTNode
from copy import copy
from utility.TypeCheckUtility import TypeCheckUtility
typeChecker=TypeCheckUtility.getInstance()

class Function(ASTNode):
  """ Encapsulate a function
      @group Accessors: set*, get*
      @sort: __init__, set*, get*
  """

  def __init__(self, name, arguments, numParen=0):
    """ Function's constructor
        @param name: Name of the function
        @param arguments: Arguments of the function
        @param numParen: Number of parenthesis around this function
        @type name: string
        @type arguments: list of L{ASTNode}
        @type numParen: int
    """
    if typeChecker.debug: 
      typeChecker.typeCheck([name,arguments], [str,list])
      for arg in arguments:
        typeChecker.typeCheck([arg],[ASTNode])

    self.__arguments=copy(arguments)
    self.__name=name
    #Number of parenthesis around
    self.__numParen=numParen
    
  def getName(self):
    """ Returns the name of this function.
        @rtype: string         
    """
    return self.__name

  def getArgs(self):
    """ Returns the arguments of this function.  Note:
        A copy of the list is returned to prevent from
        directly modifying the datastructure. 
        @rtype: list of L{ASTNode}     
    """        
    return copy(self.__arguments)
  
  def getNumParen(self):
    """ Returns the number of parenthesis around this ASTNode
        @rtype: int
    """
    return self.__numParen
        
  def setName(self,name):
    """ Sets the name of this function
        @param name: name of the function
        @type name: string
    """
    if typeChecker.debug: typeChecker.typeCheck([name], [str])
    self.__name=name
  
  def setArgs(self,arguments):
    """ Sets the arguments of this function
        @param arguments: list of arguments
        @type arguments: list of L{ASTNode}
    """
    if typeChecker.debug:
      typeChecker.typeCheck([arguments], [list])
      for arg in arguments:
        typeChecker.typeCheck([args], [ASTNode])  
    self.__arguments=copy(arguments)

  def setNumParen(self, numParen):
    """ Set the number of parenthesis around this number
        @param numParen: number of parenthesis
        @type numParen: int
    """
    self.__numParen=numParen       

  def accept(self,visitor):
    if typeChecker.debug: typeChecker.typeCheck([visitor], [ASTVisitor])
    visitor.visitFunction(self)
    
  def clone(self):
    """ Returns a copy of this object
        @rtype: L{Function}
    """ 
    newArgs=[]
    for arg in self.__arguments:
      newArgs.append(arg.clone())
    return Function(self.__name,
                    newArgs,
                    self.__numParen)

  def equals(self, anObject):
    if not (isinstance(anObject,Function) and\
            anObject.getName()==self.__name and\
            len(self.__arguments)==len(anObject.getArgs()) and\
            anObject.getNumParen()==self.__numParen):
      return 0
    myArgs=self.__arguments
    otherArgs=anObject.getArgs()
    for argIndex in xrange(len(myArgs)):
      if not myArgs[argIndex].equals(otherArgs[argIndex]):
        return 0
    return 1    