""" Range Reference node of the AST.
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca    
"""

from ASTVisitor import ASTVisitor
from ASTNode import ASTNode
from CellRef import CellRef
from utility.TypeCheckUtility import TypeCheckUtility
from utility.ErrorUtility import ErrorUtility
typeChecker=TypeCheckUtility.getInstance()
errHandler=ErrorUtility.getInstance()

class RangeRef(ASTNode):
  """ Encapsulates a reference to a range of cells
      @group Accessors: set*, get*
      @sort: __init__, set*, get*
  """

  def __init__(self, cornerA, cornerB):
    """ RangeRef's constructor
        @param cornerA: First reference of the range
        @param cornerB: Second reference of the range
        @type cornerA: L{CellRef}
        @type cornerB: L{CellRef}
    """
    if typeChecker.debug: typeChecker.typeCheck([cornerA,cornerB], 
                                                [CellRef,CellRef])
    self.__cornerA=cornerA.clone()
    self.__cornerB=cornerB.clone()
   
  def getCellRefSet(self):
    """ Returns a list of the CellRef in the
        rectangle created by cornerA and
        cornerB.  The cells are always returned
        starting from the top-left cell, down to the
        bottom-right cell.
        @rtype: list of L{CellRef}
    """
    ans=[]

    colA=min(self.__cornerA.getColumn(),self.__cornerB.getColumn())
    colB=max(self.__cornerA.getColumn(),self.__cornerB.getColumn())
    rowA=min(self.__cornerA.getRow(),self.__cornerB.getRow())
    rowB=max(self.__cornerA.getRow(),self.__cornerB.getRow())

    for i in xrange(rowA,rowB+1):
      for j in xrange(colA, colB+1):
        ans.append(CellRef(j,i,1,1))

    return ans


  def getFirstCorner(self):
    """ Returns the first corner of the range
        @rtype: L{CellRef}
    """
    return self.__cornerA.clone()

  def getSecondCorner(self):
    """ Returns the second corner of the range
        @rtype: L{CellRef}
    """
    return self.__cornerB.clone()
              
  def setFirstCorner(self, cornerA):
    """ Sets the first corner of the range
        @param cornerA: first corner
        @type cornerA: L{CellRef}
    """  
    if typeChecker.debug: typeChecker.typeCheck([cornerA], 
                                                [CellRef])
    self.__cornerA=cornerA.clone()
  
  def setSecondCorner(self, cornerB):
    """ Sets the second corner of the range
        @param cornerB: second corner
        @type cornerB: L{CellRef}
    """     
    if typeChecker.debug: typeChecker.typeCheck([cornerB], 
                                                [CellRef])  
    self.__cornerB=cornerB.clone()
       
  def accept(self, visitor):
    if typeChecker.debug: typeChecker.typeCheck([visitor], [ASTVisitor])
    visitor.visitRangeRef(self)

  def clone(self):
    """ Returns a copy of this object
        @rtype: L{RangeRef}
    """     
    return RangeRef(self.__cornerA.clone(),
                    self.__cornerB.clone())  

  def equals(self, anObject):    
    return isinstance(anObject,RangeRef) and\
           anObject.getFirstCorner().equals(self.__cornerA) and\
           anObject.getSecondCorner().equals(self.__cornerB)      