""" Testsuite for the abstract syntax tree module
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca
"""

from myast.ASTNode import ASTNode
from myast.Number import Number
from myast.CellRef import CellRef
from myast.RangeRef import RangeRef
from myast.Function import Function
from utility.TypeCheckUtility import TypeCheckUtility
from utility.ErrorUtility import ErrorUtility
from utility.ColumnConvertUtility import ColumnConvertUtility

import unittest
import sys

from math import pow
from math import fabs

from random import randint
from random import seed

TRUE=1
FALSE=0
SEED=666

MINNUM=-1000
MAXNUM=1000

seed(666)

typeChecker = TypeCheckUtility.getInstance()
errHandler  = ErrorUtility.getInstance()
converter   = ColumnConvertUtility.getInstance()

errHandler.setTesting()

class ASTTestSuite(unittest.TestCase):
    """Testing AST for success
    """

    sampleNumbers = [[5, 0, "=5"],
                     [4.5, 1, "=(4.5)"],
                     [3.2, 0, "=3.2"],
                     [-10, 3, "=(((-10)))"],
                     [32, 2, "=((32))"],
                     [12.332, 0, "=12.332"],
                     [4.32, 4, "=((((4.32))))"],
                     [-0.002, 0, "=-0.002"],
                     [11, 0, "=11"]
                    ]

    #list of tuples (cellref, column, row, isAbsoluteColumn, isAbsoluteRow)
    sampleCellRefs = [ [converter.stringToInt("A"), 4, 0, 0, 0, "=A5"],
                       [converter.stringToInt("A"), 54, 0, 0, 1, "=(A55)"],
                       [converter.stringToInt("A"), 541, 0, 0, 2, "=((A542))"],
                       [converter.stringToInt("A"), 4, 1, 0, 0, "=$A5"],
                       [converter.stringToInt("AZ"), 4, 1, 0, 3, "=((($AZ5)))"],
                       [converter.stringToInt("A"), 4, 1, 1, 0, "=$A$5"],
                       [converter.stringToInt("AD"), 4, 1, 0, 1, "=($AD5)"],
                       [converter.stringToInt("AB"), 531, 1, 1, 0, "=$AB$532"],
                       [converter.stringToInt("FD"), 299, 0, 0, 3, "=(((FD300)))"],
                       [converter.stringToInt("A"), 14, 0, 0, 0, "=A15"],
                       [converter.stringToInt("A"), 142, 0, 1, 2, "=((A$143))"]
                     ]

    sampleRangeRefs = [ [CellRef(converter.stringToInt("A"), 0, 0, 0),
                         CellRef(converter.stringToInt("B"), 5, 0, 0),
                         "=A1:B6",
                         [CellRef(converter.stringToInt("A"), 0, 1, 1),
                          CellRef(converter.stringToInt("B"), 0, 1, 1),
                          CellRef(converter.stringToInt("A"), 1, 1, 1),
                          CellRef(converter.stringToInt("B"), 1, 1, 1),
                          CellRef(converter.stringToInt("A"), 2, 1, 1),
                          CellRef(converter.stringToInt("B"), 2, 1, 1),
                          CellRef(converter.stringToInt("A"), 3, 1, 1),
                          CellRef(converter.stringToInt("B"), 3, 1, 1),
                          CellRef(converter.stringToInt("A"), 4, 1, 1),
                          CellRef(converter.stringToInt("B"), 4, 1, 1),
                          CellRef(converter.stringToInt("A"), 5, 1, 1),
                          CellRef(converter.stringToInt("B"), 5, 1, 1)
                         ]
                        ],
                        [CellRef(converter.stringToInt("B"), 5, 0, 0),
                         CellRef(converter.stringToInt("A"), 0, 0, 0),
                         "=B6:A1",
                         [CellRef(converter.stringToInt("A"), 0, 1, 1),
                          CellRef(converter.stringToInt("B"), 0, 1, 1),
                          CellRef(converter.stringToInt("A"), 1, 1, 1),
                          CellRef(converter.stringToInt("B"), 1, 1, 1),
                          CellRef(converter.stringToInt("A"), 2, 1, 1),
                          CellRef(converter.stringToInt("B"), 2, 1, 1),
                          CellRef(converter.stringToInt("A"), 3, 1, 1),
                          CellRef(converter.stringToInt("B"), 3, 1, 1),
                          CellRef(converter.stringToInt("A"), 4, 1, 1),
                          CellRef(converter.stringToInt("B"), 4, 1, 1),
                          CellRef(converter.stringToInt("A"), 5, 1, 1),
                          CellRef(converter.stringToInt("B"), 5, 1, 1)
                         ]
                        ],
                        [CellRef(converter.stringToInt("C"), 3, 0, 0),
                         CellRef(converter.stringToInt("A"), 0, 0, 0),
                         "=C4:A1",
                         [CellRef(converter.stringToInt("A"), 0, 1, 1),
                          CellRef(converter.stringToInt("B"), 0, 1, 1),
                          CellRef(converter.stringToInt("C"), 0, 1, 1),
                          CellRef(converter.stringToInt("A"), 1, 1, 1),
                          CellRef(converter.stringToInt("B"), 1, 1, 1),
                          CellRef(converter.stringToInt("C"), 1, 1, 1),
                          CellRef(converter.stringToInt("A"), 2, 1, 1),
                          CellRef(converter.stringToInt("B"), 2, 1, 1),
                          CellRef(converter.stringToInt("C"), 2, 1, 1),
                          CellRef(converter.stringToInt("A"), 3, 1, 1),
                          CellRef(converter.stringToInt("B"), 3, 1, 1),
                          CellRef(converter.stringToInt("C"), 3, 1, 1)
                         ]
                        ],
                        [CellRef(converter.stringToInt("C"), 0, 0, 0),
                         CellRef(converter.stringToInt("A"), 3, 0, 0),
                         "=C1:A4",
                         [CellRef(converter.stringToInt("A"), 0, 1, 1),
                          CellRef(converter.stringToInt("B"), 0, 1, 1),
                          CellRef(converter.stringToInt("C"), 0, 1, 1),
                          CellRef(converter.stringToInt("A"), 1, 1, 1),
                          CellRef(converter.stringToInt("B"), 1, 1, 1),
                          CellRef(converter.stringToInt("C"), 1, 1, 1),
                          CellRef(converter.stringToInt("A"), 2, 1, 1),
                          CellRef(converter.stringToInt("B"), 2, 1, 1),
                          CellRef(converter.stringToInt("C"), 2, 1, 1),
                          CellRef(converter.stringToInt("A"), 3, 1, 1),
                          CellRef(converter.stringToInt("B"), 3, 1, 1),
                          CellRef(converter.stringToInt("C"), 3, 1, 1)
                         ]
                        ]
                      ]

    sampleFunctions = [ ["f",
                         [Number(5),
                          CellRef(converter.stringToInt("A"), 5, 1, 1),
                          Number(10)
                         ],
                         0,
                         "=f(5,$A$6,10)"
                        ],
                        ["f",
                         [Function("f",
                                   [Number(10),
                                    RangeRef(CellRef(converter.stringToInt("A"), 0, 0, 0),
                                             CellRef(converter.stringToInt("A"), 5, 0, 0))
                                   ]),
                          Number(100)
                         ],
                         0,
                         "=f(f(10,A1:A6),100)"
                        ],
                        ["sum",
                         [Number(5),
                          Function("invSum", [Number(5)], 1),
                          Function("mul", [Number(5), Number(10)])
                         ],
                         0,
                         "=5-(5)+5*10"
                        ],
                        ["unaryMinus",[Number(34)],0,"=-34"
                        ],
                        ["sum",[Function("sum", [Number(23), Number(34)], 1),
                                Function("unaryMinus", [Number(45)], 1)],
                         2,
                         "=(((23+34)+(-45)))"
                        ],
                        ["sum",[Function("mul", [Number(4),
                                                 CellRef(1, 1, 0, 1)]),
                                Number(34)
                               ],
                         0,
                         "=4*B$2+34"
                        ],
                        ["mul", [Function("sum",[RangeRef(CellRef(converter.stringToInt("A"),
                                                                0,
                                                                0,
                                                                0),
                                                        CellRef(converter.stringToInt("D"),
                                                                10,
                                                                0,
                                                                0)
                                                       )
                                               ]
                                          ),
                                 Number(34),
                                 Number(45)
                                ],
                         0,
                         "=sum(A1:D11)*34*45"
                        ],
                        ["mul", [Number(34),
                                 Function("invMul",[Number(45)], 1),
                                 Function("sum", [Number(45), Number(23)], 1),
                                 Function("exp", [Number(67), Number(2)])
                                ],
                         0,
                         "=34/(45)*(45+23)*67^2"
                        ],
                        ["mul", [Number(34),
                                 Function("invMul",[Number(45)], 1),
                                 Function("sum", [CellRef(1, 1, 1, 0),
                                                  Number(23)], 1),
                                 Function("exp", [Number(67),
                                                  Number(2)])
                                ],
                         0,
                         "=34/(45)*($B2+23)*67^2"
                        ],
                        ["exp", [Function("sum",[Number(2), Number(3)]),
                                 Number(10)],
                         1,
                         "=(2+3^10)"
                        ],
                        ["exp", [Function("sum",[Number(2), Number(3)],2),
                                 Number(10)],
                         1,
                         "=(((2+3))^10)"
                        ],
                        ["unaryMinus",[Number(34,1)],
                         0,
                         "=-(34)"
                        ],
                        ["sum", [Number(3),
                                 Number(4),
                                 Function("unaryMinus", [Number(3)])],
                         0,
                         "=3+4+-3"
                        ],
                        ["sum", [Number(3),
                                 CellRef(2,2,0,0,1),
                                 Function("unaryMinus", [Number(3)])],
                         0,
                         "=3+(C3)+-3"
                        ],
                        ["sum", [Function("sum", [Number(4), Number(34)]),
                                 Number(4),
                                 Function("unaryMinus", [Number(3)])],
                         0,
                         "=4+34+4+-3"
                        ],
                        ["sum", [Number(3),
                                 Number(4.345),
                                 Function("unaryMinus", [Number(3)])],
                         0,
                         "=3+4.345+-3"
                        ],
                        ["function", [], 0, "=function()"]

                      ]

    def testNumbersGetSetValue(self):
      """Testing Number class for success (get/set Value/NumParen)
      """
      for numberTuple in self.sampleNumbers:
        value=numberTuple[0]
        numParen=numberTuple[1]
        n=Number(value, numParen)

        self.assertEqual(n.getValue(), value,
                  "getValue() method did not return the appropriate value")
        self.assertEqual(n.getNumParen(), numParen,
                  "getNumParen() method did not return the appropriate value")

        newVal=randint(MINNUM,MAXNUM)
        newPar=randint(MINNUM,MAXNUM)
        n.setValue(newVal)
        self.assertEqual(newVal, n.getValue())
        n.setNumParen(newPar)
        self.assertEqual(newPar, n.getNumParen())

    def testNumbersCloneEqual(self):
      """Testing Number class for success (clone, equals)
      """
      for numberTuple in self.sampleNumbers:
        value=numberTuple[0]
        numParen=numberTuple[1]
        n=Number(value, numParen)

        cloned=n.clone()
        self.failIf(n==cloned,
                       "Cloned instance of Number must not refer "+\
                       "to original instance")
        self.failIf(not n.equals(cloned),
                        "Number cloned instance must be equal to the original")
        otherVal=n.getValue()+randint(1,MAXNUM)
        newNumber=Number(otherVal)
        self.failIf(newNumber.equals(n),
                       "A number is equal to a number with a different value")

    def testNumberStringRepr(self):
      """Testing Number class string representation
      """
      for numberTuple in self.sampleNumbers:
        value=numberTuple[0]
        numParen=numberTuple[1]
        #expected output
        strRepr=numberTuple[2]
        n=Number(value, numParen)

        self.assertEqual(str(n),strRepr,
                         "Invalid CellRef string representation: "+str(n))

    def testCellRefAccessors(self):
      """Testing CellRef for success (get/set Row/Column/NumParen)
      """

      for cellRefTuple in self.sampleCellRefs:
        col=cellRefTuple[0]
        row=cellRefTuple[1]
        absCol=cellRefTuple[2]
        absRow=cellRefTuple[3]
        numParen=cellRefTuple[4]
        c=CellRef(col, row, absCol, absRow, numParen)

        self.failIf(c.getColumn()!=col,
                    "getColumn() method returned invalid value")
        self.failIf(c.getRow()!=row,
                    "getRow() method returned invalid value")
        self.assertEqual(c.getNumParen(),numParen,
                  "getNumParen() method did not return the appropriate value")

        newCol=randint(MINNUM,MAXNUM)
        newRow=randint(MINNUM,MAXNUM)
        newPar=randint(MINNUM,MAXNUM)
        c.setColumn(newCol)
        c.setRow(newRow)
        self.assertEqual(newCol,c.getColumn())
        self.assertEqual(newRow,c.getRow())
        c.setNumParen(newPar)
        self.assertEqual(newPar, c.getNumParen())

    def testCellRefIsAbs(self):
      """Testing CellRef for success (isAbsRow/Column)
      """

      for cellRefTuple in self.sampleCellRefs:
        col=cellRefTuple[0]
        row=cellRefTuple[1]
        absCol=cellRefTuple[2]
        absRow=cellRefTuple[3]
        numParen=cellRefTuple[4]
        c=CellRef(col, row, absCol, absRow, numParen)

        self.failIf(c.isAbsColumn()!=absCol,
                    "isAbsColumn() method returned invalid value")

        self.failIf(c.isAbsRow()!=absRow,
                    "isAbsRow() method returned invalid value")
        newIsAbsCol=randint(FALSE,TRUE)
        newIsAbsRow=randint(FALSE,TRUE)
        c.setIsAbsColumn(newIsAbsCol)
        c.setIsAbsRow(newIsAbsRow)
        self.assertEqual(newIsAbsCol,c.isAbsColumn())
        self.assertEqual(newIsAbsRow,c.isAbsRow())

    def testCellRefCloneEquals(self):
      """Testing CellRef for success (clone, equals)
      """
      for cellRefTuple in self.sampleCellRefs:
        col=cellRefTuple[0]
        row=cellRefTuple[1]
        absCol=cellRefTuple[2]
        absRow=cellRefTuple[3]
        numParen=cellRefTuple[4]
        c=CellRef(col, row, absCol, absRow, numParen)

        cloned=c.clone()
        self.failIf(c==cloned,
                      "Cloned instance of CellRef must not refer "+\
                      "to original instance")
        self.failIf(not c.equals(cloned),
                        "CellRef cloned instance must be equal to the original")

        otherColumn=c.getColumn()+randint(1,MAXNUM)
        otherRow=c.getRow()+randint(1,MAXNUM)
        otherIsAbsCol=c.isAbsColumn()+randint(1,MAXNUM)
        otherIsAbsRow=c.isAbsRow()+randint(1,MAXNUM)
        newCellRef=CellRef(otherColumn,
                           otherRow,
                           otherIsAbsCol,
                           otherIsAbsRow)
        self.failIf(newCellRef.equals(c),
                      "A cellref is equal to a cellref with a different value")

    def testCellRefString(self):
      """Testing CellRef string representation
      """
      for cellRefTuple in self.sampleCellRefs:
        col=cellRefTuple[0]
        row=cellRefTuple[1]
        absCol=cellRefTuple[2]
        absRow=cellRefTuple[3]
        numParen=cellRefTuple[4]
        #expected output
        strRepr=cellRefTuple[5]
        c=CellRef(col, row, absCol, absRow, numParen)
        self.assertEqual(str(c),strRepr,
                         "Invalid CellRef string representation: "+str(c))

    def testRangeRefAccessors(self):
      """Testing RangeRef for success (get/set cornerA/cornerB)
      """
      for rangeRefTuple in self.sampleRangeRefs:
        cornerA=rangeRefTuple[0]
        cornerB=rangeRefTuple[1]
        rangeRef=RangeRef(cornerA, cornerB)

        self.failIf(not cornerA.equals(rangeRef.getFirstCorner()),
                    "getFirstCorner returned invalid CellRef")
        self.failIf(not cornerB.equals(rangeRef.getSecondCorner()),
                    "getSecondCorner returned invalid CellRef")
        self.failIf(cornerA==rangeRef.getFirstCorner(),
                    "getFirstCorner breaks encapsulation")
        self.failIf(cornerB==rangeRef.getSecondCorner(),
                    "getSecondCorner breaks encapsulation")

        newColA=cornerA.getColumn()+randint(1, MAXNUM)
        newCornerA=cornerA.clone()
        newCornerA.setColumn(newColA)

        newColB=cornerB.getColumn()+randint(1, MAXNUM)
        newCornerB=cornerB.clone()
        newCornerB.setColumn(newColB)



        rangeRef.setFirstCorner(newCornerA)
        rangeRef.setSecondCorner(newCornerB)

        self.failIf(cornerA.equals(rangeRef.getFirstCorner()),
                    "getFirstCorner returned invalid CellRef")
        self.failIf(cornerB.equals(rangeRef.getSecondCorner()),
                    "getSecondCorner returned invalid CellRef")

    def testRangeRefCellSet(self):
      """Testing RangeRef for success (getCellRefSet)
      """
      for rangeRefTuple in self.sampleRangeRefs:
        cornerA=rangeRefTuple[0]
        cornerB=rangeRefTuple[1]
        expectedSet=rangeRefTuple[3]
        r=RangeRef(cornerA, cornerB)
        set=r.getCellRefSet()

        self.failIf(len(set)!=len(expectedSet),
              "CellRefSet is not of expected length:"+str(len(expectedSet)))

        for i in xrange(len(set)):
          self.failIf(not set[i].equals(expectedSet[i]),
            "An element in the returned CellRefSet is invalid")

      maxQty=0
      constantCellRef=self.sampleCellRefs[0]
      cornerA=CellRef(constantCellRef[1],
                      constantCellRef[2],
                      constantCellRef[3],
                      constantCellRef[4])
      for curCellRef in self.sampleCellRefs:
        cornerB=CellRef(curCellRef[1],
                        curCellRef[2],
                        curCellRef[3],
                        curCellRef[4])
        curRangeRef=RangeRef(cornerA,cornerB)

        height=abs(cornerA.getRow()-cornerB.getRow())+1
        width=abs(cornerA.getColumn()-cornerB.getColumn())+1

        self.failIf(height*width!=len(curRangeRef.getCellRefSet()),
            "Length of CellRefSet is not of expected length:"+\
            str(height*width)+"!="+str(len(curRangeRef.getCellRefSet())))

    def testRangeRefCloneEquals(self):
      """Testing RangeRef for success (clone, equals)
      """
      for rangeRefTuple in self.sampleRangeRefs:
        cornerA=rangeRefTuple[0]
        cornerB=rangeRefTuple[1]
        r=RangeRef(cornerA, cornerB)
        cloned=r.clone()
        self.failIf(r==cloned,
                      "Cloned instance of RangeRef must not refer "+\
                      "to original instance")
        self.failIf(not r.equals(cloned),
                        "RangeRef cloned instance must be equal to the original")
        cornerA=r.getFirstCorner()
        cornerB=r.getSecondCorner()
        newCornerA=CellRef(cornerA.getColumn()+randint(0,4),
                                           cornerA.getRow(),
                                           cornerA.isAbsColumn(),
                                           cornerA.isAbsRow())
        newCornerB=CellRef(cornerB.getColumn()+randint(1,10),
                                           cornerB.getRow(),
                                           cornerB.isAbsColumn(),
                                           cornerB.isAbsRow())
        newRangeRef=RangeRef(newCornerA,
                             newCornerB)
        self.failIf(newRangeRef.equals(r),
                      "A rangeref is equal to a cellref with a different value")

    def testRangeRefStringRepr(self):
      """Testing RangeRef string representation
      """
      for rangeRefTuple in self.sampleRangeRefs:
        cornerA=rangeRefTuple[0]
        cornerB=rangeRefTuple[1]
        expectedStrRepr=rangeRefTuple[2]
        r=RangeRef(cornerA, cornerB)

        self.assertEqual(str(r),expectedStrRepr,
                        "Invalid CellRef string representation: "+str(r))

    def testFunctionAccessorSuccess(self):
      """Testing Function for success (accessors)
      """
      for functionTuple in self.sampleFunctions:
        funcName=functionTuple[0]
        args=functionTuple[1]
        numParen=functionTuple[2]
        f=Function(funcName, args, numParen)

        self.failIf(f.getName()!=funcName)
        self.failIf(f.getNumParen()!=numParen)

        for i in xrange(len(f.getArgs())):
            self.failIf(not args[i].equals(f.getArgs()[i]))

        newPar=randint(MINNUM,MAXNUM)
        f.setNumParen(newPar)
        self.failIf(f.getNumParen()!=newPar)

    def testFunctionCloneEquals(self):
      """Testing Function for success (clone, equals)
      """
      for functionTuple in self.sampleFunctions:
        funcName=functionTuple[0]
        args=functionTuple[1]
        numParen=functionTuple[2]
        f=Function(funcName, args, numParen)

        cloned=f.clone()
        self.failIf(f==cloned,
                      "Cloned instance of Function must not refer "+\
                      "to original instance")
        self.failIf(not f.equals(cloned),
                        "Function cloned instance must be equal to the original")
      #assuming that every sample in the set is unique
      for funcIndex in xrange(len(self.sampleFunctions)):
        if funcIndex+1<len(self.sampleFunctions):
          f1=Function(self.sampleFunctions[funcIndex][0],
                      self.sampleFunctions[funcIndex][1])
          f2=Function(self.sampleFunctions[funcIndex+1][0],
                      self.sampleFunctions[funcIndex+1][1])
          self.failIf(f1.equals(f2),
                      "Two unequal functions were found to be equals")

    def testFunctionStrRepr(self):
      """Testing Function string representation
      """
      for functionTuple in self.sampleFunctions:
        funcName=functionTuple[0]
        args=functionTuple[1]
        numParen=functionTuple[2]
        expectedStrRepr=functionTuple[3]
        f=Function(funcName, args, numParen)

        self.assertEqual(str(f), expectedStrRepr,
                         "Incorrect string repr: "+str(f))

    def testNumberFailure(self):
        """testing Number for failure (TypeError)
        """
        self.assertRaises(TypeError,Number,"32")

    def testCellRefFailureTypeError(self):
        """testing CellRef for failure (TypeError)
        """
        self.assertRaises(TypeError, CellRef,Number(34),1,1,1)
        self.assertRaises(TypeError, CellRef,"34", 34,0,1)

    def testRangeRefFailure(self):
        """testing RangeRef for failure (TypeError)
        """
        self.assertRaises(TypeError, RangeRef, Function("min",[Number(34)]),CellRef(0,0,0,0))
        self.assertRaises(TypeError, RangeRef, CellRef(0,0,0,0), Number(34))


    def testFunctionFailure(self):
        """testing Function for failure (TypeError)
        """
        self.assertRaises(TypeError, Function, 34, [])
        self.assertRaises(TypeError, Function, "max", Number(34))



if __name__ == "__main__":
    unittest.main()
