""" Class encapsulating a cell coordinate
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca
"""

from utility.TypeCheckUtility import TypeCheckUtility
from utility.ColumnConvertUtility import ColumnConvertUtility

typeChecker = TypeCheckUtility.getInstance()
converter   = ColumnConvertUtility.getInstance()

from math import log

class CellCoord:
  """ A coordinate class (unbounded)
      @group Accessors: row*, col*
  """

  def __init__(self, col, row):
    """ CellCoord's constructor.
        @param col: Cell coordinate column
        @param row: Cell coordinate row
        @type col: int
        @type row: int
    """
    if typeChecker.debug: typeChecker.typeCheck([col, row], [int, int])
    self.__row=row
    self.__col=col

  def row(self):
    """ Returns the row of this cell coordinate
        @rtype: int
    """
    return self.__row

  def col(self):
    """ Returns the column of this cell coordinate
        @rtype: int
    """
    return self.__col

  def __hash__(self):
    """ Hash function to be used by built-in python's dictionnaries.
    """
    from SSheetData import MAXROW
    return (self.__col*int(log(MAXROW, 10)+1)+self.__row)

  def clone(self):
    """ Returns a copy of this object
        @rtype: L{CellCoord}
    """
    return CellCoord(col=self.__col,row=self.__row)

  def __lt__(self, coord):
    """ Overloading operator <
        @param coord: The compared cell coordinate
        @type coord: L{CellCoord}
    """
    if typeChecker.debug: typeChecker.typeCheck([coord], [CellCoord])
    return self.__row<coord.row() and self.__col<coord.col()

  def __le__(self, coord):
    """ Overloading operator <=
        @param coord: The compared cell coordinate
        @type coord: L{CellCoord}
    """
    if typeChecker.debug: typeChecker.typeCheck([coord], [CellCoord])
    return self.__row<=coord.row() and self.__col<=coord.col()

  def __gt__(self, coord):
    """ Overloading operator >
        @param coord: The compared cell coordinate
        @type coord: L{CellCoord}
    """
    if typeChecker.debug: typeChecker.typeCheck([coord], [CellCoord])
    return self.__row>coord.row() and self.__col>coord.col()

  def __ge__(self, coord):
    """ Overloading operator >=
        @param coord: The compared cell coordinate
        @type coord: L{CellCoord}
    """
    if typeChecker.debug: typeChecker.typeCheck([coord], [CellCoord])
    return self.__row>=coord.row() and self.__col>=coord.col()

  def __eq__(self, obj):
    """ Overloading operator ==
        @param obj: The compared cell coordinate
        @type obj: Object
    """
    if not isinstance(obj, CellCoord):
      return 0
    return self.__row==obj.row() and self.__col==obj.col()

  def __ne__(self, obj):
    """ Overloading operator !=
        @param obj: The compared cell coordinate
        @type obj: Object
    """
    if not isinstance(obj, CellCoord):
      return 1
    return self.__row!=obj.row() or self.__col!=obj.col()

  def __str__(self):
    return converter.intToString(self.__col)+str(self.__row+1)
