""" Class implementing a union (float,string)
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca    
"""

from utility.ErrorUtility import *
from utility.TypeCheckUtility import TypeCheckUtility

typeChecker = TypeCheckUtility.getInstance()
errHandler = ErrorUtility.getInstance()

class CellValue:
  """ Class implementing a union (float,string).
  """

  def __init__(self, valueFloat=None, valueString=None): 
    """ CellValue's constructor. Takes two facultative
        arguments, which cannot be specified at the same
        time.
        @param valueFloat: specify the float value
        @param valueString: specify the string value
        @type valueFloat: float
        @type valueString: string
        @requires: Not both parameters specified at the same time
    """        
    if typeChecker.debug: 
      typeChecker.typeCheck([valueFloat, valueString], 
                            [(int,float,type(None)), (str,type(None))])              
    #default vals
    self.__valueFloat=0.0 
    self.__valueString=""
    self.__isFloat=False
    self.__isString=False
    
    if valueFloat!=None and valueString!=None:
      v=ValueError("CellValue instantiated as being both float and string")
      errHandler.exceptionRaised(v)
    elif valueFloat!=None:
      self.__valueFloat=float(valueFloat)
      self.__isFloat=True
    elif valueString!=None:
      self.__valueString=valueString
      self.__isString=True
      
  def isFloat(self):
    """ Returns true if this object is a float
        @rtype: bool
    """
    return self.__isFloat

  def isString(self):
    """ Returns true if this object is a string
        @rtype: string
    """
    return self.__isString

  def getFloat(self):
    """ Returns a float value, if this object is a float.
        Otherwise, it raises an exception.
        @raises TypeError: When this object is not a float
        @rtype: float
    """
    if self.__isFloat:
      return self.__valueFloat
    else:
      v = TypeError("CellValue is not a float")
      errHandler.exceptionRaised(v)

  def getString(self):
    """ Returns a string value, if this object is a string.
        Otherwise, it raises an exception.
        @raises TypeError: When this object is not a string
        @rtype: string
    """  
    if self.__isString:
      return self.__valueString
    else:
      v = TypeError("CellValue is not a string")
      errHandler.exceptionRaised(v)

  def setFloat(self, value):
    """ Set the float value, if this object is a float.
        Otherwise, it raises an exception.
        @raises TypeError: When this object is not a float
    """   
    if typeChecker.debug: typeChecker.typeCheck([value], [(int,float)])
    if self.__isFloat:
      self.__valueFloat = float(value)
    else:
      v = TypeError("CellValue is not a float")
      errHandler.exceptionRaised(v)
  
  def setString(self, value):
    """ Set the string value, if this object is a string.
        Otherwise, it raises an exception.
        @raises TypeError: When this object is not a string
    """   
    if typeChecker.debug: typeChecker.typeCheck([value], [str])
    if self.__isString:
      self.__valueString = value
    else:
      v = TypeError("CellValue is not a string")
      errHandler.exceptionRaised(v)  

  def equals(self, obj):
    """ Returns true if obj equals this object.
        @param obj: the compared object
        @type obj: Object
    """
    if not isinstance(obj, CellValue):
      return 0
    if self.__isFloat     == obj.isFloat() and\
       self.__isString    == obj.isString():
       if obj.isFloat():
         return self.__valueFloat  == obj.getFloat()
       elif obj.isString():
         return self.__valueString == obj.getString()
       #CellValue are equals by default if not defined
       else:
         return True                    

  def __str__(self):    
    if self.__isString:
      return self.__valueString
    else:
      return str(self.__valueFloat)