""" Class that visits the AST and evaluate it.
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca
"""

from utility.ErrorUtility import *
from utility.TypeCheckUtility import TypeCheckUtility
from myast.ASTVisitor import ASTVisitor
from myast.Function import Function
from myast.Number import Number
from myast.CellRef import CellRef
from myast.RangeRef import RangeRef

from CellCoord import CellCoord
from Functions import *
from CellValue import CellValue

from types import ListType,IntType,FloatType
from string import split

errHandler  = ErrorUtility.getInstance()
typeChecker = TypeCheckUtility.getInstance()

#******************************************************************************
### class EvalVisitor
class EvalVisitor(ASTVisitor):
  """Visit the formula AST and evaluate it
  """

  def __init__(self,sheet):
    """ EvalVisitor's constructor.
        A reference to SSheetData is needed to resolve and
        evaluate references to other cells.  If a reference does
        not exist in SSheetData, it evaluate to 0. Note that it
        is assumed that cycles were handled beforehand (i.e.
        This visitor is not executed on cycles).
    """
    from SSheetData import SSheetData
    if typeChecker.debug: typeChecker.typeCheck([sheet], [SSheetData])
    self.__evalStack=[]
    self.__sheetDataSubject=sheet
    self.__valueError=False
    self.__refError=False
    self.__nameError=False
    self.__nameErrorStr=""

  def __checkValueError(self, functionArgs):
    """ Returns True if the function arguments contain strings.
        @param functionArgs: the function arguments
        @type functionArgs: list of string or int or float
        @rtype: bool
    """
    if typeChecker.debug: typeChecker.typeCheck([functionArgs], [list])
    for arg in functionArgs:
      if not (type(arg) == IntType or type(arg) == FloatType):
        self.__valueError=True
        return True
    return False

  def __checkRefError(self, cellCoord):
    """ Returns True if the cellCoord is outside of the
        spreadsheet's boundaries.
        @param cellCoord: cell coordinate
        @type cellCoord: L{CellCoord}
        @rtype: bool
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord], [CellCoord])
    sheetLimits=self.__sheetDataSubject.getLimits()
    if cellCoord.col()<sheetLimits[0].col() or\
        cellCoord.row()<sheetLimits[0].row() or\
        sheetLimits[1].col()<=cellCoord.col() or\
        sheetLimits[1].row()<=cellCoord.row():
          self.__refError=True
          return True
    return False

  def getEvaluatedValue(self):
    """ Returns the value computed by this visitor.
        @rtype: L{CellValue}
    """
    if isinstance(self.__evalStack[0], str):
      return CellValue(valueString=self.__evalStack[0])
    else:
      return CellValue(valueFloat=self.__evalStack[0])

  def hasNameError(self):
    """ Returns true if this visitor detected a name error.
        @rtype: bool
    """
    return self.__nameError

  def hasValueError(self):
    """ Returns true if this visitor detected a value error.
        @rtype: bool
    """
    return self.__valueError

  def hasRefError(self):
    """ Returns true if this visitor detected a ref error.
        @rtype: bool
    """
    return self.__refError

  def getNameErrorStr(self):
    """ Returns the invalid function name that was detected.
        @rtype: string
    """
    return self.__nameErrorStr

  def visitNumber(self, number):
    if typeChecker.debug: typeChecker.typeCheck([number], [Number])
    self.__evalStack.append(number.getValue())

  def visitFunction(self, function):
    if typeChecker.debug: typeChecker.typeCheck([function], [Function])
    numArg=0
    for argument in function.getArgs():
      if isinstance(argument, RangeRef):
        numArg += len(argument.getCellRefSet())
      else:
        numArg += 1
      argument.accept(self)

    args=self.__evalStack[-numArg:]
    self.__evalStack=self.__evalStack[0:-numArg]


    if len(args)>1 and self.__checkValueError(args):
      self.__evalStack.append(0)
      return

    execStr="answer = "+function.getName()+"("+str(args)+")"
    try:
      exec execStr
    except NameError, n:
      fName=split(n[0],"'")
      self.__nameError=True
      self.__nameErrorStr=fName[1]
      self.__evalStack.append(0)
      return
    self.__evalStack.append(answer)

  def visitCellRef(self, cellRef):
    if typeChecker.debug: typeChecker.typeCheck([cellRef], [CellRef])
    cellCoord=CellCoord(cellRef.getColumn(),cellRef.getRow())
    state=self.__sheetDataSubject.getState(cellCoord,cellCoord)

    if not self.__checkRefError(cellCoord):
      value=state.getCellValue(cellCoord)
      if value.isFloat():
        self.__evalStack.append(value.getFloat())
      elif value.isString():
        self.__evalStack.append(value.getString())
      else:
        self.__evalStack.append(0)
    else:
      self.__evalStack.append(0)

  def visitRangeRef(self, rangeRef):
    for cellRef in rangeRef.getCellRefSet():
      cellRef.accept(self)
