""" Class visiting an AST to collect a list of influencer
    cells.
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca
"""

from myast.ASTVisitor import ASTVisitor
from myast.Function import Function
from myast.Number import Number
from myast.CellRef import CellRef
from myast.RangeRef import RangeRef
from utility.ErrorUtility import *
from utility.TypeCheckUtility import TypeCheckUtility

from CellCoord import CellCoord

import copy

errHandler  = ErrorUtility.getInstance()
typeChecker = TypeCheckUtility.getInstance()

#******************************************************************************
### class InfluencerVisitor
class InfluencerVisitor(ASTVisitor):
  """Class visiting an AST to create a cell dependency list
  """

  def __init__(self):
    """ This class will go through the AST and
        add all the references to other cells
        in a list.
    """
    self.__influencers=[]

  def getInfluencers(self):
    """ Returns the list of cells influencing this object
        @rtype: list of L{CellCoord}
    """
    return copy.copy(self.__influencers)

  def visitNumber(self, number):
    """ Terminal node, do nothing
    """
    if typeChecker.debug: typeChecker.typeCheck([number], [Number])

  def visitFunction(self, function):
    """ Recursively call arguments
    """
    if typeChecker.debug: typeChecker.typeCheck([function], [Function])
    for arg in function.getArgs():
      arg.accept(self)

  def visitCellRef(self, cellRef):
    """ Add the cell reference to the list
    """
    if typeChecker.debug: typeChecker.typeCheck([cellRef], [CellRef])
    self.__influencers.append(CellCoord(cellRef.getColumn(),cellRef.getRow()))

  def visitRangeRef(self, rangeRef):
    """ Recursively visits each CellRef in the range
    """
    if typeChecker.debug: typeChecker.typeCheck([rangeRef], [RangeRef])
    for cell in rangeRef.getCellRefSet():
      cell.accept(self)
