""" Class encapsulating a cell in the spreadsheet
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca
"""


#******************************************************************************
#  SSheetCell.py --- Class encapsulating a spreadsheet cell
#
#  last modified: 08/5/04
#
#                     =====================================
#                                 Marc Provost
#                          (marc.provost@mail.mcgill.ca)
#
#
#                          McGill University, Montreal
#                     =====================================
# last modified: 15/08/2016
#                               Julien Tomezach,
#                        Master Student, ENSTA Bretagne,
#                           French engineering school
#******************************************************************************


from formulaparser.FormulaToAST import FormulaToAST
from myast.ASTNode import ASTNode
from utility.ErrorUtility import *
from utility.TypeCheckUtility import TypeCheckUtility

from CellCoord import CellCoord
from CellValue import CellValue
from InfluencerVisitor import InfluencerVisitor
from UpdatedCellsManager import UpdatedCellsManager

import copy
from string import replace

errHandler  = ErrorUtility.getInstance()
typeChecker = TypeCheckUtility.getInstance()

class SSheetCell:
  """Class encapsulating a cell in the spreadsheet
     @group Accessors: set*, get*
     @sort: __init__, set*, get*
  """

  def __init__(self, cellCoord, updatedCellsManager = None):
    if typeChecker.debug: typeChecker.typeCheck([cellCoord, updatedCellsManager], [CellCoord, UpdatedCellsManager])
    self.__coords=cellCoord
    self.__value = CellValue()
    self.__formula=""
    self.__ast=None
    self.__influencers=[]
    self.__updatedCellsManager = updatedCellsManager

  def getAST(self):
    """ Return the AST associated with this cell
        @rtype: L{ASTNode}
    """
    if self.__ast==None:
      return None
    else:
      return self.__ast.clone()

  def setAST(self, ast):
    """ Set the AST of this cell.  Will be used when cells are
        copied into other cells to avoid reparsing the same
        formula string.
        @param ast: The abstract syntax tree
        @type ast: L{ASTNode}
    """
    if typeChecker.debug: typeChecker.typeCheck([ast], [(ASTNode)])

    self.__ast=ast.clone()
    self.__formula=str(ast)
    v=InfluencerVisitor()
    self.__ast.accept(v)
    self.__influencers=v.getInfluencers()
    self.__value = CellValue(valueFloat=0.0)

  def setFormula(self, formula):
    """ Set the formula and parse it into an AST.
        @param formula: the formula
        @type formula: string
    """
    if typeChecker.debug: typeChecker.typeCheck([formula], [str])

    formulaToAST=FormulaToAST()
    astTuple=formulaToAST.parseToAST(formula+";")

    if not astTuple[0]:
      #remove ';' from the formula (not used by the caller)
      errMsg = replace(astTuple[2], ";", " ")
      return (0, errMsg)
    else:
      self.__ast=astTuple[1]

      #creating the dependence list
      v=InfluencerVisitor()
      self.__ast.accept(v)
      self.__influencers=v.getInfluencers()

      if(not (self.__formula == formula) and self.__updatedCellsManager != None):
          self.__updatedCellsManager.push(self)
      self.__formula=formula
      self.__value = CellValue(valueFloat=0.0)
      return (1, astTuple[2])

  def getFormula(self):
    """ Returns the formula
        @rtype: string
    """
    return self.__formula

  def getInfluencers(self):
    """ Return a list of CellCoord of the cells
        that influence this cell for evaluation.
        (i.e., cells that are referred by this cell)
        @rtype: list of L{CellCoord}
    """
    return copy.copy(self.__influencers)

  def getValue(self):
    """ Returns the current evaluated value of this cell.
    """
    return self.__value

  def setValue(self, value):
    """ Set the value of this cell.
        @param value: The value
        @type value: L{CellValue}
    """
    if typeChecker.debug: typeChecker.typeCheck([value], [CellValue])
    # if the value actually changed, the cell adds itself to the manager
    if(not (self.__value.equals(value)) and self.__updatedCellsManager != None):
        self.__updatedCellsManager.push(self)
    self.__value=value

  def delete(self):
      """ As the cell is going to be deleted,
          the clients need to be notified of this change
          And we need to create an other instance for updatedCellsManager
      """
      if(self.__updatedCellsManager != None):
         '''
         newCell = SSheetCell(self.__coords, self.__updatedCellsManager)
         newCell.setFormula(self.__formula)
         newCell.setValue(self.__value)
         '''
         self.__value = CellValue(None, "")
         self.__formula = ""
         self.__updatedCellsManager.push(self)

  def getCoords(self):
    """ Returns the coordinates of this cell
        @rtype: L{CellCoord}
    """
    return self.__coords
