""" This module implements the data representation of a spreadsheet.
    @sort: *ERROR_FLAG, *COL, *ROW, *VISITED
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca
"""

#  ***************************************************************************
#                                        Class encapsulating a SSheetData
#
#  Author: Marc Provost, 2003
#   Last Modifications: (June 7th 2004)
#         - Refactored / Cleaned the code
#         - SSheetState is now used when copying
#         - Updated the comments.
#         - Removed SSheetData variables used as non-persitent variables.
#              -Replaced them by OUT variables arguments in
#               method calls (see toposort)
#         - Removed named constant, added global flags
#   Modifications: (August 6th, 2004)
#         - Revised the cycle detection algorithm
#         - Now using Robert E. Tarjan's algorithm
#         - Refer to SIAM Journal on Computing, Volume 1,
#           Number 2, pp. 146-160, 1972
#         - H.V.: added cycle ID
#
#
#
#   Julien Tomezach, Master Student:
#   15/08/2016: adding the UpdatedCellsManager to keep track of the updated cells
#   that need to be sent to the observers.
#  ***************************************************************************

from string import split
from myast.ASTNode import ASTNode
from utility.ErrorUtility import *
from utility.TypeCheckUtility import TypeCheckUtility

from SSheetState import SSheetState
from CellCoord import CellCoord
from CellValue import CellValue
from CopyPasteVisitor import CopyPasteVisitor
from EvalVisitor import EvalVisitor
from SetCells import SetCells
from SubjectManager import SubjectManager
from Observer import Observer
from Subject import Subject
from UpdatedCellsManager import UpdatedCellsManager

errHandler  = ErrorUtility.getInstance()
typeChecker = TypeCheckUtility.getInstance()

VALUE_ERROR_FLAG="#VALUE!"
REF_ERROR_FLAG="#REF!"
NAME_ERROR_FLAG="#NAME!"
CYCLE_ERROR_FLAG="#CYCLE"
NOT_VISITED=0
VISITED=1

MINCOL=0
MINROW=0
MAXCOL=1000
MAXROW=1000

#Subject in the *Observer Pattern*
class SSheetData(Subject):
  """ Class encapsulating the data of a spreadsheet. SSheetData is
      the Subject in the Observer pattern and communicates with
      a set of Observers. SSheetData objects are created by
      a SubjectManager.
      @group Accessors: get*
      @group Undoable/Redoable Actions: paste, importFromFile, setState
      @sort: __init__, set*, get*
  """

  def __init__(self,
               subjectManager,
               ID,
               minCol=0,
               minRow=0,
               maxCol=5000,
               maxRow=5000):

    if typeChecker.debug:
      typeChecker.typeCheck( [subjectManager, ID, minCol, minRow, maxCol, maxRow],
                             [SubjectManager, int, int, int, int, int] )
    """ SSheetData's constructor.
        @param subjectManager: Object managing several Subjects.
        @param ID: Unique ID specified by the subject manager .
        @param minCol: smallest column of the spreadsheet
        @param minRow: smallest row of the spreadsheet
        @param maxCol: maximal column of the spreadsheet
        @param maxRow: maximal row of the spreadsheet
        @type subjectManager: L{SubjectManager}
        @type ID: int
        @type minCol: int
        @type minRow: int
        @type maxCol: int
        @type maxRow: int
        @requires: MINCOL <= minCol,maxCol <= MAXCOL and MINROW <= minRow,maxRow <= MAXROW
    """

    #Each SSheetData subject has its own unique ID <int>
    self.__ID=ID

    #Dictionnary mapping CellCoord -> SSheetCell
    # Note: Empty cells are expressed implicitely.
    # i.e. by the absence of a CellCoord
    self.__cells={}

    #Dictionnary mapping observer's ID to actual observer references
    self.__observers={}

    #A reference to the main application managing several SSheetData.
    #(DSheet.py)
    self.__subjectManager=subjectManager

    #A SSheetState variable backuping a range of cell
    self.__copyState=SSheetState()

    #  An updatedCellsManager to keep track of the updated cells
    #   that need to be sent to the observers
    self.__updatedCellsManager = UpdatedCellsManager()

    #Objectified function setCells, the only entity that modify
    #self.__cells. *Command Pattern*.  Enable easy Undo/Redo.
    #That is why we pass updatedCellsManager to it.
    self.__setCells=SetCells(self.__cells, self.__updatedCellsManager)

    #DataSet limits:
    self.__limits=( CellCoord(minCol,minRow) , CellCoord(maxCol,maxRow) )



  def getLimits(self):
    """ Returns a tuple of cell coordinates representing the boundaries
        of the spreadsheet.
        @rtype: Tuple (L{CellCoord} , L{CellCoord})
    """
    return self.__limits

  def getID(self):
    """ Returns the ID of this SSheetData
        @rtype: int
    """
    return self.__ID

  def getSubjectManager(self):
    """ Returns the subject manager managing this subject
        @rtype: L{SubjectManager}
    """
    return self.__subjectManager

  def exportToFile(self, filename):
    """ Export the contents of the spreadsheet to a file. Returns
        True on success, False otherwise.
        @param filename: The name of the file
        @type filename: string
        @rtype bool
    """
    try:
      f = open(filename, "w")
      for cellCoord in self.__cells.keys():
        formula = self.__cells[cellCoord].getFormula()
        value   = self.__cells[cellCoord].getValue()
        if formula != "":
          f.write(str(cellCoord.col())+" "+str(cellCoord.row())+" "+formula+'\n')
        elif value.isString():
          f.write(str(cellCoord.col())+" "+str(cellCoord.row())+" "+value.getString()+'\n')
      f.close()
      return True
    except Exception, e:
      f.close()
      return False

  def importFromFile(self, filename):
    """ Import the contents of filename into the spreadsheet, which is
        emptied first. Returns True on success, False otherwise.
        @param filename: The name of the file
        @rtype: bool
    """
    try:
      f = open(filename, "r")
      state = SSheetState()
      coords = []
      for line in f.readlines():
        data = split(line)
        if len(data)<3:
          return False
        col=int(data[0])
        row=int(data[1])
        coord=CellCoord(col,row)
        coords.append(coord)
        formulaOrValue=data[2]
        for string in data[3:]:
          formulaValue+=" "+string
        if len(formulaOrValue)>0:
          if formulaOrValue[0]=="=":
            state.setCellFormula(coord,formulaOrValue)
          else:
            val=CellValue(valueString=formulaOrValue)
            state.setCellValue(coord,val)
      for cellCoord in self.__cells.keys():
        if not cellCoord in coords:
          state.setCellFormula(cellCoord, "")

      errTup=self.setState(state)
      f.close()
      return errTup[0]
    except Exception, e:
      print e
      f.close()
      return False

  # *** Next follows function related to the command pattern.
  # copy, paste, undo, redo ***

  def copy(self, cellCoordA, cellCoordB):
    """ Backup the cells located within the specified cell coordinates.
        Note: The order is not important.
        @param cellCoordA: The first cell coordinate
        @param cellCoordB: The second cell coordinate
        @type cellCoordA: L{CellCoord}
        @type cellCoordB: L{CellCoord}
    """

    if typeChecker.debug:
      typeChecker.typeCheck( [cellCoordA, cellCoordB],
                             [CellCoord, CellCoord] )

    # Order the cellCoords (topLeftCorner, bottomRightCorner)
    orderedTuple           = self.__orderCellCoord(cellCoordA,
                                                   cellCoordB)
    self.__copyMinCoord    = orderedTuple[0]
    self.__copyMaxCoord    = orderedTuple[1]

    #Create a new SSheetState object and
    #update it to contain the information
    #of each cell in the range.
    #Here, we do not perform a getState(), because
    #it does not return a state containing formulas AST
    self.__copyState = SSheetState()
    cells            = self.__getCells(self.__copyMinCoord,
                                       self.__copyMaxCoord)

    for cellCoord in cells.keys():
      cellAST=cells[cellCoord].getAST()
      cellValue=cells[cellCoord].getValue()
      cellFormula=cells[cellCoord].getFormula()
      self.__copyState.setCellValue(cellCoord,cellValue)
      self.__copyState.setCellFormula(cellCoord,cellFormula)
      self.__copyState.setCellAST(cellCoord,cellAST)

  def paste(self, cellCoordA, cellCoordB):
    """ Paste the cells saved previously with a call
        to copy() into the requested range. Copy
        all cells even if rangeTuple is smaller than
        the saved range.
        Note: The order of the cell coordinates is not important
        TODO: Copy multiple times if the height/width
        of the range is a multiple of the saved height/width
        @param cellCoordA: The first cell coordinate
        @param cellCoordB: The second cell coordinate
        @type cellCoordA: L{CellCoord}
        @type cellCoordB: L{CellCoord}
    """
    if typeChecker.debug:
      typeChecker.typeCheck( [cellCoordA, cellCoordB],
                             [CellCoord, CellCoord] )

    orderedTuple = self.__orderCellCoord(cellCoordA,
                                         cellCoordB)

    #Setup the source range starting cell (top-left)
    sourceMinCol=self.__copyMinCoord.col()
    sourceMinRow=self.__copyMinCoord.row()

    #Setup the target range starting cell (top-left)
    targetCell=orderedTuple[0]
    targetMinCol=targetCell.col()
    targetMinRow=targetCell.row()

    #Use the information saved in __copyState (sourceState)
    #to create a targetState using the target coordinates.
    sourceState=self.__copyState
    targetState=SSheetState()
    for cellCoord in sourceState.getCoords():
      #Determine the current target coordinate
      curDestCol=targetMinCol+(cellCoord.col()-sourceMinCol)
      curDestRow=targetMinRow+(cellCoord.row()-sourceMinRow)
      curDestCoord=CellCoord(curDestCol,curDestRow)

      cellAST=self.__copyState.getCellAST(cellCoord)

      if cellAST!=None:
        #If there is an AST, one must update the formula
        #references to reflect the context to where it is copied.
        # *See CopyPasteVisitor*
        cellAST.accept(CopyPasteVisitor(cellCoord.col(),
                                        cellCoord.row(),
                                        curDestCol,
                                        curDestRow))

        targetState.setCellAST(curDestCoord,cellAST)
      else:
        #Otherwise, we simply copy the cell value
        cellValue=self.__copyState.getCellValue(cellCoord)
        targetState.setCellValue(curDestCoord,cellValue)
    self.setState(targetState)

  def undo(self):
    """ Undo the previous action performed, if possible
    """
    if self.canUndo():
      self.__updatedCellsManager.newUpdate()
      self.__setCells.unexecute()
      self.__evaluate()
      self.notifyObservers()

  def redo(self):
    """ Redo the most recent action performed, if possible.
    """
    if self.canRedo():
      self.__updatedCellsManager.newUpdate()
      self.__setCells.reexecute()
      self.__evaluate()
      self.notifyObservers()


  def canUndo(self):
    """ Returns True if an action can be undone
    """
    return self.__setCells.canUndo()

  def canRedo(self):
    """ Returns True if an action can be redone
    """
    return self.__setCells.canRedo()

  # *** Next follows the function related to the observer pattern
  # attach, detach, notify, getState, setState.  Note: setState also
  # use the Command pattern. ***

  def attach(self, observer):
    """ Attach the specified observer to this subject.
        It will now be notified when this subject is modified.
        @param observer: The observer to be attached
        @type observer: Observer
    """
    if typeChecker.debug: typeChecker.typeCheck( [observer], [Observer] )
    self.__observers[id(observer)]=observer
    observer.notify()

  def detach(self, observer):
    """ Detach the specified observer from this subject.
        It will no longer be notified of changes to this subject.
        @param observer: The observer to be detached
        @type observer: Observer
    """
    if typeChecker.debug: typeChecker.typeCheck( [observer], [Observer] )
    del self.__observers[id(observer)]

    #If no more observer is connected to
    #this subject, notify the manager
    if len(self.__observers.keys())==0:
      self.__subjectManager.hasNoMoreObservers(self)

  def notifyObservers(self):
    """ Notify all observers that a change has
        been performed on this subject.  The
        observers will now retrieve the state
        and update themselves.
    """
    for key in self.__observers.keys():
      self.__observers[key].notify()

  def getState(self, cellCoordA=None, cellCoordB=None):
    """ Return a SSheetState object encapsulating
        each cell contained in the requested cell range.
        If no cell coordinates are specified, return the complete state.
        Note: cells AST are NOT returned. Order is not important.
        @param cellCoordA: first conrner of the range (facultative)
        @param cellCoordB: second corner of the range (facultative)
        @type cellCoordA: L{CellCoord}
        @type cellCoordB: L{CellCoord}
    """
    if typeChecker.debug:
      typeChecker.typeCheck( [cellCoordA, cellCoordB],
                             [(CellCoord,type(None)), (CellCoord,type(None))] )

    cells = {}
    if cellCoordA==None and cellCoordB==None:
      cells = self.__cells
    elif cellCoordA!=None and cellCoordB!=None:
      cells = self.__getCells(cellCoordA, cellCoordB)
    elif cellCoordB!=None:
      cells = self.__getCells(cellCoordB, cellCoordB)
    elif cellCoordA!=None:
      cells = self.__getCells(cellCoordA, cellCoordA)

    state = SSheetState()
    for cellCoord in cells.keys():
      state.setCellValue(cellCoord, cells[cellCoord].getValue())
      state.setCellFormula(cellCoord, cells[cellCoord].getFormula())

    return state

  def setState(self, state):
    """ Update the state of this object, evaluate the formulas and
        notify the observers.
        Note: Cell Coordinates that are not present in state are
        NOT modified. To explicitely delete the data of a cell coordinate,
        it must be present in state (pointing to empty formula, value, ast).
        @param state: The state to be setted
        @type state: L{SSheetState}
    """
    if typeChecker.debug: typeChecker.typeCheck( [state], [SSheetState] )
    self.__updatedCellsManager.newUpdate()
    successTuple = self.__setCells.execute(state)

    if successTuple[0]:
      self.__evaluate()
      self.notifyObservers()
    return successTuple

  def __orderCellCoord(self, cellCoordA, cellCoordB):
    if typeChecker.debug:
      typeChecker.typeCheck( [cellCoordA, cellCoordB],
                             [CellCoord, CellCoord] )
    minCol = min(cellCoordA.col(), cellCoordB.col())
    minRow = min(cellCoordA.row(), cellCoordB.row())
    maxCol = max(cellCoordA.col(), cellCoordB.col())
    maxRow = max(cellCoordA.row(), cellCoordB.row())

    return (CellCoord(minCol, minRow),
            CellCoord(maxCol, maxRow))

  def __evaluate(self):
    """ Evaluate the cells in topological order.
    """
    #Get the strong components of the dependency graph
    strongComponents = self.__strongComponents()

    #Keeping track of cycles to mark them afterwards
    cycles = []
    for component in strongComponents:
      #component of length one: possibility of having no cycle
      #(i.e. The cell could be in a cycle with itself)
      if len(component)==1:
        cellCoord = component[0]
        cell      = self.__cells[cellCoord]
        #Check if we have a cycle of length one
        if cellCoord in cell.getInfluencers():
          cell.setValue(CellValue(valueFloat=0.0))
          cycles.append(component)
        #If the cell has a formula, evaluate it
        elif cell.getAST()!=None:
          v=EvalVisitor(self)
          cell.getAST().accept(v)

          #If an error is found during evaluation, set
          #the cell value to a special flag.
          if v.hasValueError():
            cell.setValue(CellValue(valueString=VALUE_ERROR_FLAG))
          elif v.hasNameError():
            cell.setValue(CellValue(valueString=NAME_ERROR_FLAG))
          elif v.hasRefError():
            cell.setValue(CellValue(valueString=REF_ERROR_FLAG))
          #Otherwise, simply set the value!
          else:
            cell.setValue(v.getEvaluatedValue())
      #Otherwise, cycle of length len(component)
      else:
        cycles.append(component)
        #Temporarly set the value to 0.0, so that cell outside
        #the cycle can be evaluated
        for cellCoord in component:
          cell = self.__cells[cellCoord]
          cell.setValue(CellValue(valueFloat=0.0))

    # display cycle number as error flag "#CYCLEn!"
    cycleNumber=0
    for cycle in cycles:
      cycleNumber+=1
      for cellCoord in cycle:
        cell = self.__cells[cellCoord]
        cell.setValue(CellValue(valueString=CYCLE_ERROR_FLAG+str(cycleNumber)+"!"))

  def getUpdatedCells(self, clientUpdateTime):
     return self.__updatedCellsManager.getUpdatedCells(clientUpdateTime)

  def __getCells(self, cellCoordA, cellCoordB):
    """ Returns the cell contained in this subject that are within
        the specified range.
    """
    if typeChecker.debug:
      typeChecker.typeCheck( [cellCoordA, cellCoordB],
                             [CellCoord, CellCoord] )

    orderedTuple   = self.__orderCellCoord(cellCoordA,
                                           cellCoordB)
    upLeftCoord    = orderedTuple[0]
    downRightCoord = orderedTuple[1]

    cellsDict={}
    if downRightCoord==upLeftCoord:
      #If there is only one cell, the request is performed in constant time
      if self.__cells.has_key(upLeftCoord):
        cellsDict[upLeftCoord]=self.__cells[upLeftCoord]
    else:
      #Otherwise, all cells must be scanned to find the right ones
      #TODO: This scheme use low memory,
      #but is slow for range requests.  Design desision based on
      #the rarity of range requests.  This assumption may be too strong.
      #Could be improved without modifying the datastructures by keeping
      #an ordered list of coords and performing a binary search.
      for cellCoord in self.__cells.keys():
        if cellCoord>=upLeftCoord and cellCoord<=downRightCoord:
          cellsDict[cellCoord]=self.__cells[cellCoord]
    return cellsDict

  def __topoSort(self):
    """ Return a topologically sorted list of cell coordinates
    """
    cellsLabelling = {}
    sortedList     = []

    for cellCoord in self.__cells.keys():
      cellsLabelling[cellCoord] = NOT_VISITED

    for cellCoord in self.__cells.keys():
      if cellsLabelling[cellCoord] == NOT_VISITED:
        self.__dfsSort(cellCoord,
                       cellsLabelling,
                       sortedList)
    return sortedList

  def __dfsSort(self, cellCoord, cellsLabelling, sortedList):
    """ Create the topologically sorted list by performing a
        depth first search.
        @param cellCoord: The current visited coordinate
        @param cellsLabelling: mapping keeping track of visited coordinates
        @param sortedList: IN parameter containing the sorted coordinates
        @type cellCoord: L{CellCoord}
        @type cellsLabelling: hashtable mapping L{CellCoord}
                              to L{VISITED} or L{NOT_VISITED}
        @type sortedList: list of L{CellCoord}
    """
    if typeChecker.debug:
      typeChecker.typeCheck( [cellCoord, cellsLabelling, sortedList],
                             [CellCoord, dict, list] )

    if cellsLabelling.has_key(cellCoord) and\
       cellsLabelling[cellCoord] == NOT_VISITED:
      cellsLabelling[cellCoord] = VISITED

      for influencerCoord in self.__cells[cellCoord].getInfluencers():
        self.__dfsSort(influencerCoord,
                       cellsLabelling,
                       sortedList)
      sortedList.append(cellCoord)

  def __reverseGraph(self):
    """ Return a hashtable mapping a given CellCoord
        to its dependents CellCoords.
    """
    cellsLabelling = {}
    revGraph       = {}

    for cellCoord in self.__cells.keys():
      cellsLabelling[cellCoord] = NOT_VISITED
      revGraph[cellCoord] = []

    for cellCoord in self.__cells.keys():
      if cellsLabelling[cellCoord] == NOT_VISITED:
        self.__dfsReverse(cellCoord,
                          cellsLabelling,
                          revGraph)
    return revGraph

  def __dfsReverse(self, cellCoord, cellsLabelling, revGraph):
    """ Create a mapping representing the reverse graph
        by performing a depth first search.
        @param cellCoord: The current visited coordinate
        @param cellsLabelling: mapping keeping track of visited coordinates
        @param revGraph: IN parameter mapping cell coordinates to their dependents
        @type cellCoord: L{CellCoord}
        @type cellsLabelling: hashtable mapping L{CellCoord}
                              to L{VISITED} or L{NOT_VISITED}
        @type revGraph: hashtable mapping L{CellCoord} to a list of L{CellCoord}
    """

    if typeChecker.debug:
      typeChecker.typeCheck( [cellCoord, cellsLabelling, revGraph],
                             [CellCoord, dict, dict] )

    if cellsLabelling.has_key(cellCoord) and\
       cellsLabelling[cellCoord] == NOT_VISITED:
      cellsLabelling[cellCoord] = VISITED

      for influencerCoord in self.__cells[cellCoord].getInfluencers():
        if cellsLabelling.has_key(influencerCoord):
          revGraph[influencerCoord].append(cellCoord)
          self.__dfsReverse(influencerCoord,
                            cellsLabelling,
                            revGraph)

  def __strongComponents(self):
    """ Return a list of strong components of the dependency graph.
        The strong components in topological order.
    """
    sortedList       = self.__topoSort()
    revGraph         = self.__reverseGraph()
    cellsLabelling   = {}
    strongComponents = []

    for cellCoord in self.__cells.keys():
      cellsLabelling[cellCoord] = NOT_VISITED

    sortedList.reverse()
    for cellCoord in sortedList:
      if cellsLabelling[cellCoord] == NOT_VISITED:
        component = []
        self.__dfsCollect(cellCoord,
                          cellsLabelling,
                          revGraph,
                          component)
        strongComponents.append(component)

    strongComponents.reverse()
    return strongComponents

  def __dfsCollect(self, cellCoord, cellsLabelling, revGraph, component):
    """ Collect the strong components by performing a depth first search.
        @param cellCoord: The current visited coordinate
        @param cellsLabelling: mapping keeping track of visited coordinates
        @param revGraph: mapping cell coordinates to their dependents
        @param component: IN parameter collecting a component
        @type cellCoord: L{CellCoord}
        @type cellsLabelling: hashtable mapping L{CellCoord}
                              to L{VISITED} or L{NOT_VISITED}
        @type revGraph: hashtable mapping L{CellCoord} to a list of L{CellCoord}
        @type component: list of L{CellCoord}
    """
    if typeChecker.debug:
      typeChecker.typeCheck( [cellCoord, cellsLabelling, revGraph, component],
                             [CellCoord, dict, dict, list] )

    if cellsLabelling.has_key(cellCoord) and\
       cellsLabelling[cellCoord] == NOT_VISITED:
      cellsLabelling[cellCoord] = VISITED

      for dependentCoord in revGraph[cellCoord]:
        if cellsLabelling.has_key(dependentCoord):
          self.__dfsCollect(dependentCoord,
                            cellsLabelling,
                            revGraph,
                            component)
      component.append(cellCoord)
