""" Class encapsulating a snapshot of SSheetData.
    Used as a means of communication between the
    Observers and the Subject
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca
"""

from myast.ASTNode import ASTNode
from utility.ErrorUtility import *
from utility.TypeCheckUtility import TypeCheckUtility

from CellCoord import CellCoord
from CellValue import CellValue

errHandler  = ErrorUtility.getInstance()
typeChecker = TypeCheckUtility.getInstance()

class SSheetState:
  """ Class encapsulating a snapshot of SSheetData.
      Used as a means of communication between the
      Observers and the Subject
      @group Accessors: set*, get*
  """

  def __init__(self):
    self.__cellData={}

  def hasCell(self, cellCoord):
    """ Return True if this SSheetState contains cellCoord
        @param cellCoord: cell coordinate
        @type cellCoord: L{CellCoord}
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord], [CellCoord])
    return self.__cellData.has_key(cellCoord)

  def getCoords(self):
    """ Return a list of cell coordinates that this SSheetState contains.
        @rtype: list of L{CellCoord}
    """
    return self.__cellData.keys()

  def getCellValue(self, cellCoord):
    """ Return the cell value of the specified coordinate. Return the
        empty cell value if the coordinate does not exists in this SSheetState.
        @param cellCoord: cell coordinate
        @type cellCoord: L{CellCoord}
        @rtype: L{CellValue}
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord], [CellCoord])
    if self.hasCell(cellCoord):
      return self.__cellData[cellCoord][0]
    else:
      return CellValue()

  def getCellFormula(self, cellCoord):
    """ Return the formula of the specified coordinate. Return
        the empty string if the coordinate does not exists in this SSheetState
        @param cellCoord: cell coordinate
        @type cellCoord: L{CellCoord}
        @rtype: string
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord], [CellCoord])
    if self.hasCell(cellCoord):
      return self.__cellData[cellCoord][1]
    else:
      return ""

  def getCellAST(self, cellCoord):
    """ Return the abstract syntax tree of the specified coordinate.
        Return None if the coordinate does not exists in this SSheetState.
        @param cellCoord: cell coordinate
        @type cellCoord: L{CellCoord}
        @rtype: L{ASTNode}
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord], [CellCoord])
    if self.hasCell(cellCoord):
      ast=self.__cellData[cellCoord][2]
      if ast==None:
        return ast
      else:
        return ast.clone()
    else:
      return None

  def setCellValue(self, cellCoord, value):
    """ Set the value of the specified coordinate.
        @param cellCoord: cell coordinate
        @param value: value
        @type cellCoord: L{CellCoord}
        @type value: L{CellValue}
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord, value],
                                                [CellCoord, CellValue])
    if self.hasCell(cellCoord):
      self.__cellData[cellCoord][0]=value
    else:
      self.__cellData[cellCoord]=[value,"",None]

  def setCellFormula(self, cellCoord, formula):
    """ Set the formula of the specified coordinate.
        @param cellCoord: cell coordinate
        @param formula: formula
        @type cellCoord: L{CellCoord}
        @type formula: string
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord, formula],
                                                [CellCoord, str])
    if self.hasCell(cellCoord):
      self.__cellData[cellCoord][1]=formula
    else:
      self.__cellData[cellCoord]=[CellValue(),formula,None]

  def setCellAST(self, cellCoord, ast):
    """ Set the ast of the specified coordinate.
        @param cellCoord: cell coordinate
        @param ast: abstract syntax tree
        @type cellCoord: L{CellCoord}
        @type ast: L{ASTNode}
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord, ast],
                                                [CellCoord, (ASTNode, type(None))])
    if self.hasCell(cellCoord):
      self.__cellData[cellCoord][2]=ast
    else:
      self.__cellData[cellCoord]=[CellValue(),"",ast]

  def removeCell(self, cellCoord):
    """ Remove the cell coordinate, if it already exists
        @param cellCoord: cell coordinate
        @type cellCoord: L{CellCoord}
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord], [CellCoord])
    if self.hasCell(cellCoord):
      del self.__cellData[cellCoord]

  def equals(self, obj):
    """ Returns true if this object equals obj
        @param obj: Object to be compared
        @type obj: Object
        @rtype: bool
    """
    if not isinstance(obj, SSheetState):
      return False
    if not len(obj.getCoords())==len(self.getCoords()):
      return False
    for coord in self.getCoords():
      if not obj.hasCell(coord):
        return False

      myVal      = self.getCellValue(coord)
      myFormula  = self.getCellFormula(coord)
      myAST      = self.getCellAST(coord)

      objVal     = obj.getCellValue(coord)
      objFormula = obj.getCellFormula(coord)
      objAST     = obj.getCellAST(coord)

      if not myVal.equals(objVal):
        return False
      elif not myFormula==objFormula:
        return False
      elif myAST==None:
        if not objAST==None:
          return False
      else:
        if not myAST.equals(objAST):
          return False

    return True

  def clone(self):
    """ Returns an exact copy of this object
        @rtype: L{SSheetState}
    """
    cloneState=SSheetState()
    for coord in self.getCoords():
      cellValue=self.getCellValue(coord)
      cellFormula=self.getCellFormula(coord)
      cellAST=self.getCellAST(coord)
      cloneState.setCellValue(coord,cellValue)
      cloneState.setCellFormula(coord,cellFormula)
      cloneState.setCellAST(coord,cellAST)
    return cloneState
