""" Class objectifying the method that modifies the data structure of
    SSheetData.  Supports Undo/Redo.
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca
"""

#  ***************************************************************************
#                            SetCells Command
#
#  Author: Marc Provost, 2003
#  Modifications: (June 6th 2004)
#         - Refactored / Cleaned the code
#         - Added several private functions to clarify the code
#         - Renamed the variables to clearer names
#         - Updated the comments.
#   Last Modifications: Julien Tomezach, Master Student.
#           _add UpdatedCellsManager
#  ***************************************************************************

from utility.TypeCheckUtility import TypeCheckUtility
from utility.ErrorUtility import ErrorUtility
from utility.ErrorUtility import LOG
from utility.ErrorUtility import WARNING
from utility.ErrorUtility import FATAL

from CellValue import CellValue
from SSheetCell import SSheetCell
from Command import Command
from CellCoord import CellCoord
from SSheetState import SSheetState
from UpdatedCellsManager import UpdatedCellsManager

typeChecker=TypeCheckUtility.getInstance()
errHandler=ErrorUtility.getInstance()

class SetCells(Command):
  """ Command that execute all the actions that modify
      the dictionnary holding the cells of a SSheetData object.
  """

  def __init__(self, cells, updatedCellsManager = None):
    """ SetCells's constructor. Takes a reference to the dictionnary
        holding the cells of a SSheetData object.
        @param cells: Dictionnary holding cell information.
        @type cells: hashtable mapping L{CellCoord} to L{SSheetCell}
    """
    if typeChecker.debug: typeChecker.typeCheck([cells, updatedCellsManager], [dict, UpdatedCellsManager])
    #cells is the receiver : execute/unexecute are performed on it.
    self.__cells=cells
    #history of performed actions: 2-D List Of Tuples Of SSheetCell
    # [ ActionA:[ (cellBefore1, cellAfter1), (cellBefore2, cellAfter2), ... ],
    #   ActionB:[...],
    #   ...,
    # ]
    self.__historyOfCells=[]
    self.__historyIndex=-1
    self.__updatedCellsManager = updatedCellsManager;

  #*** Next follows private accessors to the receiver __cells

  def __hasCell(self, cellCoord):
    """ Returns true if __cells has a key cellCoord
        @param cellCoord: cell coordinate
        @type cellCoord: L{CellCoord}
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord], [CellCoord])
    return self.__cells.has_key(cellCoord)

  def __getCell(self, cellCoord):
    """ Return the cell mapped by cellCoord. Returns an empty cell
        if cellCoord is not present in __cells.
        @param cellCoord: cell coordinate
        @type cellCoord: L{CellCoord}
        @rtype: L{SSheetCell}
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord], [CellCoord])
    if self.__hasCell(cellCoord):
      return self.__cells[cellCoord]
    else:
      return SSheetCell(cellCoord, self.__updatedCellsManager)

  def __getCellFormula(self, cellCoord):
    """ Return the formula of the cell located at cellCoord.
        Returns an empty string if cellCoord is not present in __cells.
        @param cellCoord: cell coordinate
        @type cellCoord: L{CellCoord}
        @rtype: string
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord], [CellCoord])
    if self.__hasCell(cellCoord):
      return self.__cells[cellCoord].getFormula()
    else:
      return ""

  def __getCellValue(self, cellCoord):
    """ Return the value of the cell located at cellCoord.
        Returns an empty value if cellCoord is not present in __cells.
        @param cellCoord: cell coordinate
        @type cellCoord: L{CellCoord}
        @rtype: L{CellValue}
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord], [CellCoord])
    if self.__hasCell(cellCoord):
      return self.__cells[cellCoord].getValue()
    else:
      return CellValue()

  def __getCellAST(self, cellCoord):
    """ Return the ast of the cell located at cellCoord. Returns None if
        cellCoord is not present in __cells.
        @param cellCoord: cell coordinate
        @type cellCoord: L{CellCoord}
        @rtype: L{CellValue}
    """
    if typeChecker.debug: typeChecker.typeCheck([cellCoord], [CellCoord])
    if self.__hasCell(cellCoord):
      return self.__cells[cellCoord].getAST()
    else:
      return None

  def __putCell(self, cell):
    """ Put cell into __cells. If cell is empty, delete
        the corresponding cell located in __cells.
        @param cell: cell
        @type cell: L{SSheetCell}
    """
    if typeChecker.debug: typeChecker.typeCheck([cell], [SSheetCell])
    formulaStr=cell.getFormula()
    ast=cell.getAST()
    value=cell.getValue()
    cellCoord=cell.getCoords()
    if formulaStr=="" and ast==None and (not value.isString()):
      # If cell is emtpy, it is expressed implicitely by not being there
      if self.__hasCell(cellCoord):
        print "SetCells: we delete an empty cell"
        self.__cells[cellCoord].delete()
        del self.__cells[cellCoord]
    else:
      # Otherwise, simply add cell to __cells
      self.__cells[cellCoord]=cell

  def __areCurCellsAffected(self, nextCells):
    """ Before applying an action on __cells, verify
        that there is at least one cell that is modified.
        To prevent from undoing/redoing an action that changed nothing!
        @param nextCells: list of the cells that will be putted in __cells
        @type nextCells: list of L{SSheetCell}
        @rtype: bool
    """
    if typeChecker.debug:
      typeChecker.typeCheck([nextCells], [list])
      for nextCell in nextCells:
        typeChecker.typeCheck([nextCell], [SSheetCell])

    affected=0
    for nextCell in nextCells:
      nextFormulaStr=nextCell.getFormula()
      nextAST=nextCell.getAST()
      nextValue=nextCell.getValue()

      curFormulaStr=self.__getCellFormula(nextCell.getCoords())
      curAST=self.__getCellAST(nextCell.getCoords())
      curValue=self.__getCellValue(nextCell.getCoords())

      if nextFormulaStr=="" and nextAST==None and (not nextValue.isString()):
        if curFormulaStr!="":
          affected=1
        if curValue.isString():
          affected=1
      else:
        if curFormulaStr!=nextFormulaStr:
          affected=1
        elif curValue.isString() and nextValue.isString():
          if curValue.getString()!=nextValue.getString():
            affected=1
        elif nextValue.isString():
          affected=1

    return affected


  def __backupAffectedCells(self, nextCells):
    """ Remember an action to be able to undo/redo it.
        @param nextCells: cells that will be putted in __cells
        @type nextCells: list of L{SSheetCell}
    """
    if typeChecker.debug:
      typeChecker.typeCheck([nextCells], [list])
      for nextCell in nextCells:
        typeChecker.typeCheck([nextCell], [SSheetCell])

    curAndNextCells=[]

    for nextCell in nextCells:
      curCell=self.__getCell(nextCell.getCoords())
      curAndNextCells.append((curCell,nextCell))

    self.__historyOfCells=self.__historyOfCells[:self.__historyIndex+1]+\
                          [curAndNextCells]+\
                          self.__historyOfCells[self.__historyIndex+1:]

  def __retrievePreviousCells(self):
    """ Get the cells of the previous action (to undo)
        @rtype: list of L{SSheetCell}
    """
    previousCells=[]
    for cellTuple in self.__historyOfCells[self.__historyIndex]:
      previousCells.append(cellTuple[0])
    return previousCells

  def __retrieveNextCells(self):
    """ Get the cells of the next action (to redo)
        @rtype: list of L{SSheetCell}
    """
    nextCells=[]
    for cellTuple in self.__historyOfCells[self.__historyIndex+1]:
      nextCells.append(cellTuple[1])
    return nextCells

  def __setPreviousCellsAsCurrent(self):
    """ Set the previous state (in history) as being current.
        (used just after undo)
    """
    self.__historyIndex-=1

  def __setNextCellsAsCurrent(self):
    """ Set the next state (in history) as being current.
        (used just after redo)
    """
    self.__historyIndex+=1

  def __createCells(self, state):
    """ Transform a SSheetState into an equivalent
        list of SSheetCell. If the list Of SSheetCells is None,
        creating the cells failed due to a parse error.
        errMsg describe the parse error.
        @param state: The state to be transformed into cells
        @type state: L{SSheetState}
        @rtype: tuple (list of L{SSheetCell}, string)
    """
    if typeChecker.debug: typeChecker.typeCheck([state], [SSheetState])

    cells=[]

    for cellCoord in state.getCoords():
      value=state.getCellValue(cellCoord)
      formulaStr=state.getCellFormula(cellCoord)
      ast=state.getCellAST(cellCoord)
      cell=SSheetCell(cellCoord, self.__updatedCellsManager)

      #Here to avoid reparsing, we first check if there is already an AST
      if ast != None:
        cell.setAST(ast)
      #We reparse if there is a formula string
      elif formulaStr!="":
        parseTuple=cell.setFormula(formulaStr)
        #If there is a parse error, it is propagated
        if not parseTuple[0]:
          return (None, parseTuple[1])
      #If the value is a non-empty string
      elif value.isString() and value.getString()!="":
        cell.setValue(value)
      #Otherwise, the cell is empty and will be deleted from cellData

      cells.append(cell)

    return (cells, "")

  def execute(self, nextState):
    """ Update __cells with the data in nextState.
        Return a tuple to propagate parse errors.
        The first element of the tuple is true if
        the execute was a success, false otherwise.
        The second element is the error message.
        @param nextState: state that updates __cells
        @type nextState: L{SSheetState}
        @rtype: tuple (bool, string)
    """
    if typeChecker.debug: typeChecker.typeCheck([nextState], [SSheetState])

    nextCellsTuple = self.__createCells(nextState)
    if nextCellsTuple[0]==None:#parse error
      return (False, nextCellsTuple[1])
    elif self.__areCurCellsAffected(nextCellsTuple[0]):
      self.__backupAffectedCells(nextCellsTuple[0])

      for cell in nextCellsTuple[0]:
        self.__putCell(cell)

      self.__setNextCellsAsCurrent()

    return (True, nextCellsTuple[1])


  def unexecute(self):
    """ Undo the last action.
    """
    if self.canUndo():
      previousCells=self.__retrievePreviousCells()

      for cell in previousCells:
        self.__putCell(cell)

      self.__setPreviousCellsAsCurrent()


  def reexecute(self):
    """ Redo the next action.
    """
    if self.canRedo():
      nextCells=self.__retrieveNextCells()

      for cell in nextCells:
        self.__putCell(cell)

      self.__setNextCellsAsCurrent()

  def canUndo(self):
    """ Returns true if it is possible to undo
        @rtype: bool
    """
    return self.__historyIndex>=0 and self.__historyIndex<len(self.__historyOfCells)

  def canRedo(self):
    """ Returns true if it is possible to redo
        @rtype: bool
    """
    return self.__historyIndex>=-1 and self.__historyIndex<len(self.__historyOfCells)-1
