""" Testsuite for the subject data module
    @author: Marc Provost
    @organization: McGill University
    @license: GNU General Public License
    @contact: marc.provost@mail.mcgill.ca
"""

from formulaparser.FormulaParser import FormulaParser
from formulaparser.FormulaToAST import FormulaToAST
from myast.ASTVisitor import ASTVisitor
from myast.Number import Number
from myast.CellRef import CellRef
from myast.RangeRef import RangeRef
from myast.Function import Function

from subjectdata.SSheetCell import SSheetCell
from subjectdata.CellValue import CellValue
from subjectdata.CellCoord import CellCoord
from subjectdata.InfluencerVisitor import InfluencerVisitor
from subjectdata.SSheetState import SSheetState
from subjectdata.Functions import *
from subjectdata.SSheetData import SSheetData
from subjectdata.SSheetData import VALUE_ERROR_FLAG
from subjectdata.SSheetData import REF_ERROR_FLAG
from subjectdata.SSheetData import NAME_ERROR_FLAG
from subjectdata.SSheetData import CYCLE_ERROR_FLAG
from subjectdata.SubjectManager import SubjectManager

import unittest
import sys
from math import pow
from math import fabs
from random import randint

from utility.ErrorUtility import ErrorUtility
from utility.TypeCheckUtility import TypeCheckUtility
from utility.ColumnConvertUtility import ColumnConvertUtility

errHandler  = ErrorUtility.getInstance()
typeChecker = TypeCheckUtility.getInstance()
converter   = ColumnConvertUtility.getInstance()

errHandler.setTesting()

class CellCoordTestSuite(unittest.TestCase):

  sampleSuccess = [ (0,0) , (1,0) , (0,1) , (1,1) ]


  def testSSheetCellSuccess(self):
    """ Testing CellCoord for success
    """
    for sample in self.sampleSuccess:
      coord = CellCoord(sample[0], sample[1])
      self.assertEqual(sample[0], coord.col())
      self.assertEqual(sample[1], coord.row())

    #verifying overloaded operators
    self.failIf(not CellCoord(0,0)<=CellCoord(1,0))
    self.failIf(not CellCoord(0,0)<=CellCoord(0,1))
    self.failIf(not CellCoord(1,0)>=CellCoord(0,0))
    self.failIf(not CellCoord(0,1)>=CellCoord(0,0))
    self.failIf(not CellCoord(0,0)<CellCoord(1,1))
    self.failIf(not CellCoord(1,1)>CellCoord(0,0))
    self.failIf(not CellCoord(0,0)==CellCoord(0,0))
    self.failIf(not CellCoord(1,1)==CellCoord(1,1))

    #testing clone
    cell=CellCoord(1,1)
    self.failIf(not cell.clone()==cell)

  def testSSheetCellFailure(self):
    """ Testing CellCoord for failure
    """
    self.assertRaises(TypeError, CellCoord, "as", 1)
    self.assertRaises(TypeError, CellCoord, 1, "sa")
    self.assertRaises(TypeError, CellCoord, 1.0, 1)
    self.assertRaises(TypeError, CellCoord, 1.0, 1.1)
    self.assertRaises(TypeError, CellCoord(0,0).__ge__, "df")
    self.assertRaises(TypeError, CellCoord(0,0).__le__, 3.4)
    self.assertRaises(TypeError, CellCoord(0,0).__gt__, 34)
    self.assertRaises(TypeError, CellCoord(0,0).__lt__, 34)

class CellValueTestSuite(unittest.TestCase):

  sampleSuccess = [ (None, None, False, False),
                    (None, "marc", False, True),
                    (35.0, None, True, False),
                    (None, "roger", False, True),
                    (None, "rogers", False, True) ]

  def testCellValueSuccess(self):
    """ Testing CellValue for success
    """
    for i in xrange(len(self.sampleSuccess)):
      sample = self.sampleSuccess[i]
      c = CellValue(sample[0],
                    sample[1])

      self.failIf(not c.equals(c))
      #assuming all sample are unique
      anotherSample = self.sampleSuccess[(i+1)%len(self.sampleSuccess)]
      c2 = CellValue(anotherSample[0],
                     anotherSample[1])

      self.failIf(c.equals(c2))
      self.assertEqual(c.isFloat(), sample[2])
      self.assertEqual(c.isString(), sample[3])
      if c.isFloat():
        self.assertEqual(c.getFloat(), sample[0])
        c.setFloat(sample[0]+1)
        self.assertEqual(c.getFloat(), sample[0]+1)
      if c.isString():
        self.assertEqual(c.getString(), sample[1])
        c.setString(sample[1]+"a")
        self.assertEqual(c.getString(), sample[1]+"a")




  def testCellValueFailure(self):
    """ Testing CellValue for failure
    """
    self.assertRaises(ValueError, CellValue, 0.0, "")

    floatValue=CellValue(valueFloat=0.0)
    stringValue=CellValue(valueString="")
    noneValue=CellValue()
    self.assertRaises(TypeError, floatValue.setString, "df")
    self.assertRaises(TypeError, floatValue.getString)
    self.assertRaises(TypeError, stringValue.setFloat, 0.0)
    self.assertRaises(TypeError, stringValue.getFloat)
    self.assertRaises(TypeError, noneValue.setString, "df")
    self.assertRaises(TypeError, noneValue.setFloat, 34.3)
    self.assertRaises(TypeError, noneValue.getString)
    self.assertRaises(TypeError, noneValue.getFloat)

    self.assertRaises(TypeError, CellValue, "df", "")
    self.assertRaises(TypeError, CellValue, 0.0, 4.5)

    self.assertRaises(TypeError, floatValue.setFloat, "df")
    self.assertRaises(TypeError, stringValue.setString, 34.4)

class InfluencerVisitorTestSuite(unittest.TestCase):

  sampleSuccess = [ ("=A1;", [CellCoord(0,0)]),
                    ("=A1+A2;", [CellCoord(0,0), CellCoord(0,1)]),
                    ("=MIN(A1:A5);", [CellCoord(0,0),
                                     CellCoord(0,1),
                                     CellCoord(0,2),
                                     CellCoord(0,3),
                                     CellCoord(0,4)]),
                    ("=MIN(A5:A1);", [CellCoord(0,0),
                                     CellCoord(0,1),
                                     CellCoord(0,2),
                                     CellCoord(0,3),
                                     CellCoord(0,4)]),
                    ("=MIN(A5:A5);", [CellCoord(0,4)])
                  ]

  def testInfluencerVisitorSuccess(self):
    """ Testing InfluencerVisitor for success
    """
    parser = FormulaToAST()
    for sample in self.sampleSuccess:
      v        = InfluencerVisitor()
      astTuple = parser.parseToAST(sample[0])
      ast      = astTuple[1]

      #important problem if formula did not parse
      self.failIf(not astTuple[0],
                  "Parser did not recognize this formula: "+sample[0])

      ast.accept(v)
      depend   = v.getInfluencers()

      self.failIf(not len(depend)==len(sample[1]))
      #each CellCoord must be located in the expected output
      for cellCoordA in depend:
        check = False
        for cellCoordB in sample[1]:
          if cellCoordA==cellCoordB:
            check = True
        self.failIf(not check)

class SSheetCellTestSuite(unittest.TestCase):
  #Reusing formulas from previous testsuite
  formulaSuccessSamples = [("=((4));", "=((4))"),
                            ("=4+3+min(A1:A5);", "=4+3+min(A1:A5)"),
                            ("=4;", "=4"),
                            ("=sum(3,4,A1);", "=3+4+A1"),
                            ("=f(A1:A4,34)+g();", "=f(A1:A4,34)+g()"),
                            ("=(3+2+3)^(2+3);", "=(3+2+3)^(2+3)"),
                            ("=((23+A1+(34))+23);", "=((23+A1+(34))+23)"),
                            ("=(A1)+((sum(3,4)));", "=(A1)+((3+4))"),
                            ("=23^2*(3+2)^min(3,4);", "=23^2*(3+2)^min(3,4)"),
                            ("=3^(sum(3,4));", "=3^(3+4)"),
                            ("=r(A1:C456);","=r(A1:C456)"),
                            ("=(((((A1)))))+5;", "=(((((A1)))))+5"),
                            ("=(((f())));", "=(((f())))"),
                            ("=((3+4^2--3));", "=((3+4^2--3))"),
                            ("=((-(4)));", "=((-(4)))"),
                            ("=f(3,4,3+4+(5));", "=f(3,4,3+4+(5))"),
                            ("=3.45+((3.0));","=3.45+((3))"),
                            ("=3.45*(3.23+3^2+-2);","=3.45*(3.23+3^2+-2)"),
                            ("=f(3.45,2.33,2,g(A1:A5));",
                             "=f(3.45,2.33,2,g(A1:A5))"),
                            ("=f(sum(3,4,3));", "=f(3+4+3)"),
                            ("=(3/(4))+3*3;", "=(3/(4))+3*3")
                          ]
  formulaFailureSamples = ["=",
                            "34",
                            "=3+4",
                            "=A1:A5;",
                            "=f(A1:A5+4);",
                            "=--5;",
                            "=zxy;",
                            "=AA;",
                            "=56B;",
                            "=A+B;",
                            "=5!;",
                            "=34+2z;",
                            "=(3));",
                            "=f(4));",
                            "=((4+5);",
                            "=f(3,4,);",
                            "=f(A1:A2,3,A2:A4,);",
                            "=5^^4;",
                            "=3++3;",
                            "=2**3;",
                            "=f(A1:);",
                            "=A23+3,34;"
                          ]

  def testSSheetCellSuccess(self):
    """ Testing SSheetCell for success
    """
    parser = FormulaToAST()
    performanceCounter=5
    for sample in self.formulaSuccessSamples:
      astTuple  = parser.parseToAST(sample[0])

      self.failIf(not astTuple[0],
                  "Parser did not recognize this formula: "+sample[0])

      ast = astTuple[1]
      dependV = InfluencerVisitor()
      ast.accept(dependV)
      coords=CellCoord(randint(0,2000),randint(0,2000))

      cell = SSheetCell(coords)
      successTuple = cell.setFormula(sample[1])

      self.failIf(not successTuple[0])
      self.failIf(not ast.equals(cell.getAST()))
      self.assertEqual(cell.getCoords(), coords)
      self.assertEqual(sample[1], cell.getFormula())
      self.assertEqual(len(dependV.getInfluencers()), len(cell.getInfluencers()))

      #slow, do it only X times
      if performanceCounter>0:
        for cellCoordA in dependV.getInfluencers():
          check=False
          for cellCoordB in cell.getInfluencers():
            if cellCoordA==cellCoordB:
              check=True
          self.failIf(not check)
        performanceCounter-=1

      floatVal = CellValue(valueFloat=45.0)
      stringVal = CellValue(valueString="marc")
      noneVal = CellValue()

      #Value should be default float since a formula was setted
      self.failIf(not cell.getValue().equals(CellValue(valueFloat=0.0)))

      cell.setValue(floatVal)
      self.failIf(not cell.getValue().equals(CellValue(valueFloat=45.0)))
      cell.setValue(stringVal)
      self.failIf(not cell.getValue().equals(CellValue(valueString="marc")))

  def testSSheetCellFailure(self):
    """ Testing SSheetCell for failure
    """
    self.assertRaises(TypeError, SSheetCell, "df")

    c=SSheetCell(CellCoord(0,0))

    self.assertRaises(TypeError, c.setFormula, 34)
    self.assertRaises(TypeError, c.setValue, 34)
    self.assertRaises(TypeError, c.setAST, "df")

    for sample in self.formulaFailureSamples:
      successTuple = c.setFormula(sample[0])
      self.failIf(successTuple[0])

class SSheetStateTestSuite(unittest.TestCase):

  parser = FormulaToAST()
  successSamples = [ (CellCoord(0,0),
                      CellValue(valueFloat=45.0),
                      "",
                      parser.parseToAST("=4+3+min(A1:A5);")[1]),

                      (CellCoord(1,0),
                       CellValue(valueString="marc"),
                       "=f(3,4,3+4+(5));",
                       None),

                      (CellCoord(1,1),
                       CellValue(),
                       "",
                       None),

                      (CellCoord(2,1),
                       CellValue(valueFloat=4.0),
                       "=4;",
                       None)
                   ]

  def testSSheetStateSuccess(self):
    """ Testing SSheetState for success
    """
    state        = SSheetState()
    addedSamples = []

    for sample in self.successSamples:
      cellCoord = sample[0]
      value     = sample[1]
      formula   = sample[2]
      ast       = sample[3]

      state.setCellValue(cellCoord, value)
      state.setCellFormula(cellCoord, formula)
      state.setCellAST(cellCoord, ast)

      addedSamples.append(sample)
      for addedSample in addedSamples:
        cellCoord = sample[0]
        value     = sample[1]
        formula   = sample[2]
        ast       = sample[3]

        self.failIf(not value.equals(state.getCellValue(cellCoord)))
        if ast==None:
          self.failIf(not state.getCellAST(cellCoord)==None)
        else:
          self.failIf(not ast.equals(state.getCellAST(cellCoord)))
        self.assertEqual(formula, state.getCellFormula(cellCoord))

    #testing defaults
    state = SSheetState()

    state.setCellFormula(CellCoord(0,0), "=4;")
    self.failIf(not state.getCellAST(CellCoord(0,0))==None)
    self.failIf(not state.getCellValue(CellCoord(0,0)).equals(CellValue()))

    state.setCellAST(CellCoord(1,1), self.parser.parseToAST("=4+3+min(A1:A5);")[1])
    self.failIf(not state.getCellValue(CellCoord(1,1)).equals(CellValue()))
    self.failIf(not state.getCellFormula(CellCoord(1,1))=="")

  def testSSheetStateFailure(self):
    """ Testing SSheetState for failure
    """
    state = SSheetState()

    self.assertRaises(TypeError, state.setCellFormula, "sd", "")
    self.assertRaises(TypeError, state.setCellAST, "sd", "")
    self.assertRaises(TypeError, state.setCellValue, "df", "")
    self.assertRaises(TypeError, state.setCellFormula, CellCoord(0,0), 34)
    self.assertRaises(TypeError, state.setCellValue, CellCoord(0,0), 34)
    self.assertRaises(TypeError, state.setCellAST, CellCoord(0,0), "")

class FunctionsTestSuite(unittest.TestCase):

  sampleSumSuccess = [ ([4.5, 4,0, -34, -3.4], -28.9),
                       ([4.5], 4.5),
                       ([-4], -4.0)
                     ]

  sampleMulSuccess = [ ([4.5, 4.0, -34, -3.4], 2080.8),
                       ([4.5], 4.5),
                       ([-4], -4.0)
                     ]

  sampleExpSuccess = [ ([4.5, 4.0, -34, -3.4], 1.1103665490119934e+302),
                       ([2, 2], 4),
                       ([2, -1], 0.5),
                       ([4.5], 4.5),
                       ([-4], -4.0)
                     ]

  sampleInvSumSuccess = [ ([2], -2),
                          ([4.5], -4.5),
                          ([-4], 4.0)
                        ]

  sampleInvMulSuccess = [ ([2], 0.5),
                          ([4.5], 1.0/4.5),
                          ([-4], 1.0/-4.0)
                        ]




  def testFunctionsSuccess(self):
    """ Testing Functions for success
    """
    pass

  def testSumSuccess(self):
    """ Testing Sum for success
    """
    for sample in self.sampleSumSuccess:
      ans=sum(sample[0])
      self.assertAlmostEqual(ans, sample[1])

  def testMulSuccess(self):
    """ Testing Mul for success
    """
    for sample in self.sampleMulSuccess:
      ans=mul(sample[0])
      self.assertAlmostEqual(ans, sample[1])

  def testExpSuccess(self):
    """ Testing Exp for success
    """
    for sample in self.sampleExpSuccess:
      ans=exp(sample[0])
      self.assertAlmostEqual(ans, sample[1])

  def testInvSumSuccess(self):
    """ Testing InvSum for success
    """
    for sample in self.sampleInvSumSuccess:
      ans=invSum(sample[0])
      self.assertAlmostEqual(ans, sample[1])

  def testInvMulSuccess(self):
    """ Testing InvMul for success
    """
    for sample in self.sampleInvMulSuccess:
      ans=invMul(sample[0])
      self.assertAlmostEqual(ans, sample[1])

  def testSumFailure(self):
    """ Testing Sum for failure
    """
    self.assertRaises(TypeError, sum, 34,34)
    self.assertRaises(TypeError, sum, ["de"])
    self.assertRaises(TypeError, sum, None)
    self.assertRaises(ValueError, sum, [])

  def testMulFailure(self):
    """ Testing Mul for failure
    """
    self.assertRaises(TypeError, mul, 34,34)
    self.assertRaises(TypeError, mul, ["de"])
    self.assertRaises(TypeError, mul, None)
    self.assertRaises(ValueError, mul, [])

  def testExpFailure(self):
    """ Testing Exp for failure
    """
    self.assertRaises(TypeError, exp, 34,34)
    self.assertRaises(TypeError, exp, ["de"])
    self.assertRaises(TypeError, exp, None)
    self.assertRaises(ValueError, exp, [])

  def testInvSumFailure(self):
    """ Testing InvSum for failure
    """
    self.assertRaises(TypeError, invSum, 34,34)
    self.assertRaises(TypeError, invSum, ["de"])
    self.assertRaises(TypeError, invSum, None)
    self.assertRaises(ValueError, invSum, [])
    self.assertRaises(ValueError, invSum, [34,34])

  def testInvMulFailure(self):
    """ Testing InvMul for failure
    """
    self.assertRaises(TypeError, invMul, 34,34)
    self.assertRaises(TypeError, invMul, ["de"])
    self.assertRaises(TypeError, invMul, None)
    self.assertRaises(ValueError, invMul, [])
    self.assertRaises(ValueError, invMul, [34,34])

class SSheetDataTestSuite(unittest.TestCase):

  sampleSetStateSuccess = [ [ (CellCoord(0,0), "=4", CellValue(valueFloat=4.0)),
                              (CellCoord(1,0), "=3+4", CellValue(valueFloat=7.0)),
                              (CellCoord(2,0), "=sum(3,4,2)", CellValue(valueFloat=9.0)),
                              (CellCoord(2,1), "=3*3-2^2", CellValue(valueFloat=5.0))
                            ],
                            [ (CellCoord(0,0), "=2*3+1+Z34*2", CellValue(valueFloat=7.0))
                            ],
                            [ (CellCoord(10,0), "marco", CellValue(valueString="marco"))
                            ],
                            [ (CellCoord(5,5), "=A1+34", CellValue(valueFloat=41.0)),
                              (CellCoord(5,6), "=K1", CellValue(valueString="marco"))
                            ],
                            [ (CellCoord(5,5), "=A1+36*2", CellValue(valueFloat=79.0)),
                              (CellCoord(5,6), "=E92+1", CellValue(valueFloat=1.0))
                            ],
                            [ (CellCoord(22,22), "=sum(A1:C1)", CellValue(valueFloat=23.0)),
                              (CellCoord(22,23), "=sum(C1:A1)", CellValue(valueFloat=23.0))
                            ],
                            [ (CellCoord(22,24), "=K1+23", CellValue(valueString=VALUE_ERROR_FLAG)),
                              (CellCoord(22,25), "=sum(C1:A1)", CellValue(valueFloat=23.0))
                            ],
                            [ (CellCoord(22,24), "=1/A1", CellValue(valueFloat=1.0/7.0)),
                              (CellCoord(22,25), "=A1*A1", CellValue(valueFloat=7.0*7.0)),
                              (CellCoord(22,26), "=1/A1^2", CellValue(valueFloat=1/(7.0*7.0))),
                              (CellCoord(22,27), "=sum(C1:C2)*2", CellValue(valueFloat=28.0)),
                              (CellCoord(22,28), "=rogers(A1:A8)", CellValue(valueString=NAME_ERROR_FLAG)),
                              (CellCoord(22,29), "=3+(3*2-1*(3+2))", CellValue(valueFloat=4.0))
                            ],
                            [ (CellCoord(10,0), "=sum(E92:E99)", CellValue(valueFloat=0.0))
                            ],
                            [ (CellCoord(10,0), "", CellValue())
                            ]
                          ]

  sampleCopyPasteSuccess = [ (CellCoord(0,0), "=40", CellValue(valueFloat=40.0)),
                             (CellCoord(1,0), "=20", CellValue(valueFloat=20.0)),
                             (CellCoord(2,0), "=30", CellValue(valueFloat=30.0)),
                             (CellCoord(0,1), "=5", CellValue(valueFloat=5.0)),
                             (CellCoord(1,1), "=10", CellValue(valueFloat=10.0)),
                             (CellCoord(2,1), "=3", CellValue(valueFloat=3.0)),
                             (CellCoord(0,2), "=100", CellValue(valueFloat=100.0)),
                             (CellCoord(1,2), "=150", CellValue(valueFloat=150.0)),
                             (CellCoord(2,2), "=50", CellValue(valueFloat=50.0)),
                             (CellCoord(0,3), "=sum(A1:A3)/3", CellValue(valueFloat=30.0)),
                             (CellCoord(3,0), "=sum(A$1:C$1)", CellValue(valueFloat=145.0))
                           ]




  def testCopyPasteSuccess(self):
    """ Testing Copy/Paste with undo/redo for Success
    """

    state_0   = SSheetState()
    subject = SSheetData(SubjectManager(),0)
    for sample in self.sampleCopyPasteSuccess:
      state_0.setCellFormula(sample[0], sample[1])
    subject.setState(state_0)
    #get the state with evaluated formulas
    state_1=subject.getState()

    #verifying that relative references are updated
    subject.copy(CellCoord(0,3), CellCoord(0,3))
    subject.paste(CellCoord(1,3), CellCoord(1,3))
    subject.paste(CellCoord(2,3), CellCoord(2,3))

    state_2 = subject.getState()
    self.assertEqual(state_2.getCellFormula(CellCoord(1,3)),
                     "=sum(B1:B3)/3")
    self.assertEqual(state_2.getCellFormula(CellCoord(2,3)),
                     "=sum(C1:C3)/3")

    #little success/sanity check for undo/redo
    subject.undo()
    subject.undo()
    self.failIf(not state_1.equals(subject.getState()))
    subject.redo()
    subject.redo()
    self.failIf(not state_2.equals(subject.getState()))

    #Checking that absolute cells are not updated
    subject.copy(CellCoord(3,0), CellCoord(3,0))
    subject.paste(CellCoord(3,1), CellCoord(3,1))
    subject.paste(CellCoord(3,2), CellCoord(3,2))

    state_3 = subject.getState()
    self.assertEqual(state_3.getCellFormula(CellCoord(3,1)),
                     "=sum(A$1:C$1)")
    self.assertEqual(state_3.getCellFormula(CellCoord(3,2)),
                     "=sum(A$1:C$1)")

    #another little success/sanity check for undo/redo
    subject.undo()
    subject.undo()
    self.failIf(not state_2.equals(subject.getState()))
    subject.undo()
    subject.undo()
    self.failIf(not state_1.equals(subject.getState()))
    subject.redo()
    subject.redo()
    self.failIf(not state_2.equals(subject.getState()))
    subject.redo()
    subject.redo()
    self.failIf(not state_3.equals(subject.getState()))

    #testing copying a group of cells
    subject.copy(CellCoord(1,1), CellCoord(3,2))
    subject.paste(CellCoord(10,10), CellCoord(10,10))
    state_4 = subject.getState()

    for i in xrange(1,3):
      for j in xrange(1,3):
        self.assertEqual(state_4.getCellFormula(CellCoord(i,j)),
                         state_4.getCellFormula(CellCoord(i+9,j+9)))

    self.assertEqual(state_4.getCellFormula(CellCoord(12,10)), "=sum(J$1:L$1)")
    self.assertEqual(state_4.getCellFormula(CellCoord(12,11)), "=sum(J$1:L$1)")

    #trying to copy a string value
    state_5 = SSheetState()
    state_5.setCellValue(CellCoord(5,5), CellValue(valueString="marc"))
    subject.setState(state_5)

    subject.copy(CellCoord(5,5), CellCoord(5,5))
    subject.paste(CellCoord(6,6), CellCoord(6,6))
    state_6 = subject.getState()

    self.failIf(not state_6.getCellValue(CellCoord(5,5)).equals(\
                    state_6.getCellValue(CellCoord(6,6))))
    valueStrA = state_6.getCellValue(CellCoord(5,5)).getString()
    valueStrB = state_6.getCellValue(CellCoord(6,6)).getString()
    self.assertEqual(valueStrA, "marc")
    self.assertEqual(valueStrB, "marc")
    self.assertEqual(state_6.getCellFormula(CellCoord(5,5)),
                     state_6.getCellFormula(CellCoord(6,6)))
    formulaA = state_6.getCellFormula(CellCoord(5,5))
    formulaB = state_6.getCellFormula(CellCoord(6,6))
    self.assertEqual(formulaA,"")
    self.assertEqual(formulaB,"")

    #testing REF error
    subject.copy(CellCoord(0,3), CellCoord(0,3))
    subject.paste(CellCoord(5,0), CellCoord(5,0))
    state_7 = subject.getState()

    self.assertEqual(state_7.getCellValue(CellCoord(5,0)).getString(),
                     REF_ERROR_FLAG)


  def testSetStateSuccess(self):
    """ Testing SetState/evaluate with undo/redo for Success
    """
    subject = SSheetData(SubjectManager(),0)
    previousState = SSheetState()
    for sample in self.sampleSetStateSuccess:
      state   = SSheetState()
      for cellData in sample:
        if len(cellData[1])==0 or cellData[1][0]=="=":
          state.setCellFormula(cellData[0], cellData[1])
        else:
          state.setCellValue(cellData[0], CellValue(valueString=cellData[1]))
      subject.setState(state)

      retrievedState = subject.getState()
      for cellData in sample:
        sampleCoord = cellData[0]
        expectedVal = cellData[2]
        retrievedVal = retrievedState.getCellValue(sampleCoord)
        if expectedVal.isFloat():
          self.failIf(not retrievedVal.isFloat())
          self.assertAlmostEqual(retrievedVal.getFloat(),
                                 expectedVal.getFloat())
          self.failIf(not (sampleCoord in retrievedState.getCoords()))
        elif expectedVal.isString():
          self.failIf(not retrievedVal.isString())
          self.assertEqual(retrievedVal.getString(),
                           retrievedVal.getString())
          self.failIf(not (sampleCoord in retrievedState.getCoords()))
        else:
          self.failIf(retrievedVal.isString())
          self.failIf(retrievedVal.isFloat())
          self.failIf(not expectedVal.equals(retrievedVal))
          self.failIf(sampleCoord in retrievedState.getCoords())


      subject.undo()
      self.failIf(not subject.getState().equals(previousState))
      subject.redo()
      previousState = subject.getState()

  def testA(self):
    """ Testing SSheetData for Success
    """
    pass

if __name__ == "__main__":
    unittest.main()
